from fnmatch import fnmatchcase
from typing import List, Any, Callable, Set, Union
from collections.abc import Collection
import inspect

from cosapp.ports.enum import PortType, CommonPorts
from cosapp.ports.port import ExtensiblePort, Port
from cosapp.utils.helpers import check_arg


def natural_varname(name: str) -> str:
    """
    Strip references to 'inwards' and 'outwards' ports from variable name
    """
    inwards = CommonPorts.INWARDS.value
    outwards = CommonPorts.OUTWARDS.value
    name = name.strip()
    name = name.replace(f"{inwards}.", "")
    name = name.replace(f"{outwards}.", "")
    return name


def make_wishlist(wishlist: Union[str, List[str]], name="wishlist") -> List[str]:
    ok = True
    if isinstance(wishlist, str):
        wishlist = [wishlist]
    elif isinstance(wishlist, Collection):
        ok = len(wishlist) == 0 or all(isinstance(item, str) for item in wishlist)
    elif wishlist is None:
        wishlist = []
    else:
        ok = False
    if not ok:
        raise TypeError(
            f"{name!r} must be a string, or a sequence of strings; got {wishlist}."
        )
    # Filter out 'inwards' and 'outwards' in wishlist
    filtered = set(natural_varname(name) for name in wishlist)
    return list(filtered)


def get_attributes(obj) -> Set[str]:
    """Returns non-callable members of an object"""
    return set(
        m[0] for m in inspect.getmembers(obj)
        if not m[0].startswith("_") and not callable(m[1])
    )


def find_variables(
    watched_object: "cosapp.systems.System",
    includes: List[str],
    excludes: List[str],
    advanced_filter: Callable[[Any], bool] = lambda x: True,
    inputs: bool = True,
    outputs: bool = True,
    include_const: bool = False,
) -> List[str]:
    """Generate the list of requested variables.

    The variables are sorted in alphabetical order.

    Parameters
    ----------
    watched_object : cosapp.systems.System
        Object that owns the variables searched
    includes : list of str
        Variables matching these patterns will be included
    excludes : list of str
        Variables matching these patterns will be excluded
    advanced_filter : Callable[[Any], bool]
        Function taking the variable as input and returning an acceptance criteria (True if variable is valid)
    inputs : bool
        Defines if input variables will be accepted or not
    outputs : bool
        Defines if output variables will be accepted or not
    include_const : bool
        Defines if read-only properties defined with `System.add_property` will be accepted or not

    Returns
    -------
    List[str]
        Variable names matching the includes/excludes patterns of the user in the watched object.

    .. note::
        Inward and outward variables will appear without the prefix `inwards.` or `outwards.`.
    """
    from cosapp.systems.system import System, IterativeConnector  # Local import to avoid recursion
    check_arg(watched_object, 'watched_object', System)

    if not (inputs or outputs):
        # quick return, if possible
        return []

    includes = make_wishlist(includes)
    excludes = make_wishlist(excludes)
    result = set()

    def is_valid(port):
        return isinstance(port, ExtensiblePort) and (
            (port.direction is PortType.IN and inputs) or
            (port.direction is PortType.OUT and outputs)
        )

    def find_matches(name: str, value: Any) -> Set[str]:
        result = set()
        if advanced_filter(value):
            for pattern in includes:
                if fnmatchcase(name, pattern):
                    include = True
                    for pattern in excludes:
                        if fnmatchcase(name, pattern):
                            include = False
                            break
                    if include:
                        result.add(name)
                        break
        return result

    ports = set()
    names = set()

    for name, ref in watched_object.name2variable.items():
        # Save variable for non virtual port and skip IterativeConnector as open_loops append before
        # Driver._precompute
        # Suppress duplicates INWARDS and OUTWARDS
        port = ref.mapping
        valid = (
            is_valid(port)
            and name == natural_varname(name)
            and not isinstance(port.owner, IterativeConnector)
        )
        if not valid:
            continue  # skip `name`

        ports.add(port)
        names.add(name)
        result |= find_matches(name, watched_object[name])
    
    # Find matches among port properties
    Port_cls_attr = get_attributes(Port)

    for port in ports:
        if not isinstance(port, Port):
            continue  # exclude `inwards` and `outwards` extensible ports
        port_properties = get_attributes(port) - Port_cls_attr - set(port._variables)
        port_name = port.full_name(trim_root=True)
        for name in port_properties:
            result |= find_matches(f"{port_name}.{name}", getattr(port, name))

    # Find matches among system properties
    system_properties = find_system_properties(watched_object, include_const)
    for name in system_properties:
        try:
            child, attr = name.rsplit('.', maxsplit=1)
        except ValueError:
            owner, attr = watched_object, name
        else:
            owner = watched_object[child]
        result |= find_matches(name, getattr(owner, attr))

    return sorted(result)


def find_system_properties(system, include_const=False):
    """
    Returns system properties, defined either as class properties
    (with @property decorator), or with `System.add_property`.
    The latter are excluded if optional argument `include_const`
    is False (default).
    """
    from cosapp.systems.system import System  # Local import to avoid recursion
    check_arg(system, 'system', System)

    base_cls_attr = get_attributes(System)

    def local_properties(system) -> set:
        local_attr = get_attributes(system) - base_cls_attr
        if not include_const:
            local_attr -= set(system.properties)
        return local_attr

    def find_properties(system, head=None):
        properties = set()
        if head is None:
            head = system
        prefix = "" if system is head else f"{head.get_path_to_child(system)}."
        properties |= set(f"{prefix}{attr}" for attr in local_properties(system))
        for child in system.children.values():
            properties |= find_properties(child, head)
        return properties

    return find_properties(system)



import logging
from copy import deepcopy
from typing import Any, Dict, Optional, Union

import numpy

from cosapp.core.numerics.basics import MathematicalProblem
from cosapp.core.numerics.boundary import Boundary
from cosapp.drivers.driver import Driver

logger = logging.getLogger(__name__)


class RunOnce(Driver):
    """Driver running the model on its `System` owner.

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """

    __slots__ = ('initial_values', 'solution')

    def __init__(
        self, name: str, owner: "Optional[cosapp.systems.System]" = None, **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)
        
        self.initial_values = dict()  # type: Dict[str, float]
            # desc="Initial guess for the iteratives."
        self.solution = dict()  # type: Dict[str, float]
            # desc="Dictionary (name, value) of the latest solution reached."

    def set_init(self, modifications: Dict[str, Any]) -> None:
        """Define initial values for one or more variables.

        The variable can be contextual `child1.port2.var`. The only rule is that it should belong to
        the owner `System` of this driver or any of its descendants.

        Parameters
        ----------
        modifications : Dict[str, Any]
            Dictionary of (variable name, value)

        Examples
        --------
        >>> driver.set_init({'myvar': 42, 'dummy': 'banana'})
        """
        if self.owner is None:
            raise AttributeError(
                f"Driver {self.name!r} must be attached to a System to be assigned initial values."
            )

        if not isinstance(modifications, dict):
            raise TypeError(
                "Initial values must be specified through a dictionary (variable name: value)."
            )

        for variable, value in modifications.items():
            boundary = Boundary(self.owner, variable, default=value)

            if boundary.name in self.initial_values:
                old_boundary = self.initial_values[boundary.name]
                old_boundary.set_default_value(boundary.default_value, boundary.mask)
                self.initial_values[boundary.name] = old_boundary
            else:
                self.initial_values[boundary.name] = boundary

            # Set the System with the init value - useful if this driver is not inside a solver
            self.initial_values[boundary.name].set_to_default()

            # Setting a new initial value implies, we will use the init and so the solution is cleared
            self.solution.clear()

    def get_init(self, force_init: bool = False) -> numpy.ndarray:
        """Get the System iteratives initial values for this driver.

        Parameters
        ----------
        force_init : bool, optional
            Force the initial values to be used. Default is False.

        Returns
        -------
        numpy.ndarray
            The list of iteratives initial values.
            The values should be in the same order as the unknowns in the `get_problem`.
        """
        full_init = numpy.empty(0)

        for name, unknown in self.get_problem().unknowns.items():
            if not force_init and name in self.solution:
                # We ran successfully at least once and are environmental friendly
                data = self.solution[name]
            else:  # User wants the init or first simulation or crash
                if name.startswith(self.name):
                    name = name[len(self.name) + 1 : -1]

                if name in self.initial_values:
                    boundary = self.initial_values[name]
                    umask = unknown.mask if unknown.mask is not None else numpy.empty(0)
                    bmask = boundary.mask if boundary.mask is not None else numpy.empty(0)
                    if not numpy.array_equal(umask, bmask):
                        raise ValueError(
                            f"Unknown and initial conditions on {unknown.name!r} are not masked equally"
                        )
                    data = deepcopy(boundary.default_value)
                    self.initial_values[name] = boundary
                else:
                    data = deepcopy(unknown.value)

            full_init = numpy.append(full_init, data)

        return full_init

    def get_problem(self) -> MathematicalProblem:
        """Returns the full mathematical for the case.

        Returns
        -------
        MathematicalProblem
            The full mathematical problem to solve for the case
        """
        return self.owner.get_unsolved_problem()

    def setup_run(self):
        """Method called once before starting any simulation."""
        if not self.owner.is_standalone() and self.owner.parent is None:
            self.owner.open_loops()  # Force loops opening to test if the owner needs a solver

            if self.get_problem().shape != (0, 0):
                logger.warning(
                    "Required iterations detected, not taken into account in {} driver.".format(
                        type(self).__qualname__
                    )
                )

            self.owner.close_loops()

    def _precompute(self):
        """Set execution order and start the recorder."""
        super()._precompute()
        # Solution cannot be cleared in setup_run otherwise it won't be available when get_init is called.
        self.solution.clear()

    def compute(self) -> None:
        """Execute drivers on all child `System` belonging to the driver `System` owner.
        """
        if len(self.children) == 0:
            self.owner.run_children_drivers()

        if self._recorder is not None:
            self._recorder.record_state(self.name)

    def _postcompute(self):
        """Actions performed after the `Module.compute` call."""
        # Should be called in _postcompute and not clean_run otherwise it won't work for multi-points cases
        self.solution = dict(
            [
                (key, deepcopy(unknown.value))
                for key, unknown in self.get_problem().unknowns.items()
            ]
        )
        super()._postcompute()



"""Test export to FMU"""
import importlib
import itertools
import pathlib
import re
import uuid
import zipfile
from xml.etree.ElementTree import parse, tostring

import numpy
import pandas
import pytest
from numpy.testing import assert_almost_equal

from cosapp.drivers import MonteCarlo, NonLinearSolver, RungeKutta, RunSingleCase
from cosapp.ports.enum import PortType
from cosapp.ports.port import ExtensiblePort
from cosapp.recorders import DataFrameRecorder
from cosapp.systems import System
from cosapp.tools.fmu.exporter import (
    Fmi2Causality,
    Fmi2Variability,
    FmuBuilder,
    TimeIntegrator,
    VariableType,
    to_fmu,
)

# Check optional third parties availability
pytest.importorskip("jinja2")
pytest.importorskip("pythonfmu")


def load_fmu_class(module_name: str, filepath: str):
    # Import the user interface
    spec = importlib.util.spec_from_file_location(module_name, filepath)
    fmu_interface = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(fmu_interface)
    # Instantiate the interface
    with open(filepath, "r") as f:
        data = f.read()
        class_name = re.search("class (\w+)\(\s*Fmi2Slave\s*\)\s*:", data).group(1)
    return getattr(fmu_interface, class_name)(instance_name="instance")


def get_variable_names(vars):
    return [var.name for var in vars]


@pytest.mark.parametrize("value", TimeIntegrator)
def test_TimeIntegrator_driver(value):
    assert (
        value.driver == "RungeKutta"
        if value.value.startswith("Runge-Kutta")
        else value.name
    )


@pytest.mark.parametrize("value", TimeIntegrator)
def test_TimeIntegrator_options(value):
    expected = (
        dict(order=int(value.value.split()[1]))
        if value.value.startswith("Runge-Kutta")
        else dict()
    )
    assert value.options == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (True, VariableType.Boolean),
        (2, VariableType.Integer),
        (42.0, VariableType.Real),
        ("hello", VariableType.String),
        (b"hello", TypeError),
        (object(), TypeError),
        (numpy.array(True), VariableType.Boolean),
        (numpy.array(2), VariableType.Integer),
        (numpy.array(42.0), VariableType.Real),
        (numpy.array("hello", dtype="U"), VariableType.String),
        (numpy.array([1, 2]), TypeError),
        (numpy.array(b"hello", dtype="S"), TypeError),
    ],
)
def test_FMUBuilder__get_variable_type(value, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            FmuBuilder._get_variable_type(value)
    else:
        assert FmuBuilder._get_variable_type(value) == expected


@pytest.mark.parametrize(
    "causality, variability",
    itertools.product(Fmi2Causality, (list(Fmi2Variability) + [None])),
)
def test_FMUBuilder__add_variables(causality, variability):
    data = dict((("a", True), ("b", 2), ("c", 42.0), ("d", "hello")))

    vars = FmuBuilder._add_variables(data, causality, variability)

    for v in vars:
        assert v.name in data
        assert v.causality == causality.name
        if variability is None:
            assert v.variability is None
        else:
            assert v.variability == variability.name
    assert len(vars) == len(data)

    with pytest.raises(TypeError):
        FmuBuilder._add_variables({"data": b"hello"}, causality, variability)


def test_FMUBuilder__get_default_value(testtype):
    variables = FmuBuilder._get_default_value(
        list(testtype.inputs[System.INWARDS]), testtype
    )

    for k, v in variables.items():
        if k == "k":
            assert_almost_equal(v, getattr(testtype, k))
        else:
            assert v == getattr(testtype, k)

    variables = FmuBuilder._get_default_value(("k[0]", "k[1]", "k[2]"), testtype)

    for i in range(3):
        assert variables[f"k[{i}]"] == testtype["k"][i]

    with pytest.raises(ValueError):
        FmuBuilder._get_default_value(("k[4]",), testtype)

    with pytest.raises(ValueError):
        FmuBuilder._get_default_value(("rr",), testtype)


def test_FMUBuilder__get_default_variables():
    p1 = ExtensiblePort(System.INWARDS, PortType.IN)
    p1.add_variable("a", True)
    p1.add_variable("b", 2)
    p1.add_variable("c", 42.0)
    p1.add_variable("d", "hello")
    p1.add_variable("e", b"hello")

    p2 = ExtensiblePort("out", PortType.OUT)
    p2.add_variable("f", object())
    p2.add_variable("g", numpy.array(True))
    p2.add_variable("h", numpy.array(2))
    p2.add_variable("i", numpy.array(42.0))
    p2.add_variable("j", numpy.array("hello", dtype="U"))
    p2.add_variable("k", numpy.array([1, 2]))
    p2.add_variable("l", numpy.array(b"hello", dtype="S"))

    out = FmuBuilder._get_default_variables({"p1": p1, System.OUTWARDS: p2}, to_skip=())
    expected = ("p1.a", "p1.b", "p1.c", "p1.d", "g", "h", "i", "j")
    for e in expected:
        assert e in out
    assert len(out) == len(expected)

    out = FmuBuilder._get_default_variables(
        {"p1": p1, System.OUTWARDS: p2}, to_skip=["p1.c", "p1.a", "j"]
    )
    expected = ("p1.b", "p1.d", "g", "h", "i")
    for e in expected:
        assert e in out
    assert len(out) == len(expected)


def test_FMU_export_all_causality(tmp_path, allcausality):
    test = allcausality
    fmu_file = to_fmu(test, dest=tmp_path)

    if fmu_file.suffix == ".fmu":
        with zipfile.ZipFile(fmu_file, "r") as zip_desc:
            zip_desc.extractall(fmu_file.parent)

    fmu_slave = load_fmu_class(
        "allcausality", str(fmu_file.parent / "resources" / "allcausality.py")
    )

    assert len(fmu_slave.vars) == 4
    names = get_variable_names(fmu_slave.vars.values())
    assert "in_.x" in names
    assert fmu_slave.vars[names.index("in_.x")].causality == Fmi2Causality.input
    assert "in_x" in names
    assert fmu_slave.vars[names.index("in_x")].causality == Fmi2Causality.parameter
    assert "out.x" in names
    assert fmu_slave.vars[names.index("out.x")].causality == Fmi2Causality.output
    assert "out_y" in names
    assert fmu_slave.vars[names.index("out_y")].causality == Fmi2Causality.local


def test_FMU_export_hide_all_variables(tmp_path, allcausality):
    test = allcausality
    fmu_folder = FmuBuilder.generate_fmu_facade(
        test, dest=tmp_path, inputs=[], locals=[], outputs=[], parameters=[]
    )

    fmu_slave = load_fmu_class(
        "allcausality",
        str(FmuBuilder._get_project_folder(tmp_path) / "allcausality.py"),
    )

    assert len(fmu_slave.vars) == 0


def test_FMU_export_all_type(tmp_path, testtype):
    test = testtype
    fmu_folder = FmuBuilder.generate_fmu_facade(test, dest=tmp_path)

    fmu_slave = load_fmu_class(
        "testtype", str(FmuBuilder._get_project_folder(tmp_path) / "testtype.py")
    )

    assert len(fmu_slave.vars) == 8


@pytest.mark.parametrize("name", ["k", "l"])
def test_FMU_export_type_failure(tmp_path, testtype, name):
    test = testtype
    with pytest.raises(TypeError):
        to_fmu(test, dest=tmp_path, parameters={name: testtype[name]})


def test_FMU_export_non_linear_multipoints(tmp_path, ode):
    solver = NonLinearSolver("solver")
    solver.add_child(RunSingleCase("pt1"))
    solver.add_child(RunSingleCase("pt2"))

    with pytest.raises(ValueError):
        to_fmu(ode, dest=tmp_path, non_linear_solver=solver)


def test_FMU_export_non_linear_bad_child(tmp_path, ode):
    solver = NonLinearSolver("solver")
    solver.add_child(MonteCarlo("dummy"))

    with pytest.raises(TypeError):
        to_fmu(ode, dest=tmp_path, non_linear_solver=solver)


def test_FMU_handle_vector_variable(tmp_path, vector_syst):
    pyfmi = pytest.importorskip("pyfmi")

    fmu_file = to_fmu(
        vector_syst,
        dest=tmp_path,
        inputs=["vi[0]", "vr[1]"],
        outputs=["vi[2]", "vr[2]"],
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    variables = model.get_model_variables()

    assert model.get_integer([variables["vi[0]"].value_reference])[0] == 1
    model.set_integer([variables["vi[2]"].value_reference], [33])
    assert model.get_integer([variables["vi[2]"].value_reference])[0] == 33

    assert model.get_real([variables["vr[2]"].value_reference])[0] == 3.0
    model.set_real([variables["vr[1]"].value_reference], [42.0])
    assert model.get_real([variables["vr[1]"].value_reference])[0] == 42.0



@pytest.mark.parametrize("integrator", TimeIntegrator)
def test_FMU_export_time_integrator(tmp_path, ode, integrator):
    pyfmi = pytest.importorskip("pyfmi")
    fmu_file = to_fmu(
        ode,
        dest=tmp_path,
        locals={"f": 0.0, "df_dt": 0.0},
        time_integrator=integrator,
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    # Run the FMU simulation
    results = model.simulate(final_time=1)

    # Reference value
    end_time = results["time"][-1]
    ref = ode(end_time)
    assert results["f"][-1] == pytest.approx(ref, rel=1e-3)



def test_FMU_integration_simple(tmp_path, ode):
    pyfmi = pytest.importorskip("pyfmi")

    # Convert to FMU
    fmu_file = to_fmu(
        ode,
        dest=tmp_path,
        locals=["f", "df_dt"],
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    # Run the FMU simulation
    results = model.simulate(final_time=1)

    # Reference value
    end_time = results["time"][-1]
    ref = ode(end_time)
    assert results["f"][-1] == pytest.approx(ref, rel=1e-7)


def test_FMU_integration_nonlinear(tmp_path, iterativenonlinear):
    pyfmi = pytest.importorskip("pyfmi")

    # Setup the system
    iterativenonlinear.splitter.split_ratio = 0.1
    iterativenonlinear.mult2.K1 = 1
    iterativenonlinear.mult2.K2 = 1
    iterativenonlinear.nonlinear.k1 = 1
    iterativenonlinear.nonlinear.k2 = 0.5
    iterativenonlinear.p_in.x = 1.0

    solver = NonLinearSolver("solver", iterativenonlinear)
    solver.runner.offdesign.add_unknown("nonlinear.k1").add_equation(
        "splitter.p2_out.x == 10."
    )

    # Init
    t_driver = iterativenonlinear.add_driver(
        RungeKutta(order=4, dt=0.1, time_interval=[0, 0.2])
    )
    t_driver.add_child(solver)

    # Set BC
    t_driver.set_scenario(
        init={"p_in.x": 1.0}, values={"p_in.x": "5. - 4. * exp(- t / 0.1)"}
    )
    recorder = t_driver.add_recorder(
        DataFrameRecorder(includes=["p_in.x", "p_out.x", "nonlinear.k1"]), period=0.1
    )
    iterativenonlinear.run_drivers()

    # Reference
    data = recorder.export_data()
    assert_almost_equal(data["p_out.x"], numpy.full(len(data), 10.0))

    # Convert to FMU
    fmu_file = to_fmu(
        iterativenonlinear,
        dest=tmp_path,
        non_linear_solver=solver,
        locals=["nonlinear.k1"],
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    # Run the FMU simulation
    inputs = ("p_in.x", lambda t: 5.0 - 4.0 * numpy.exp(-t / 0.1))
    results = model.simulate(final_time=0.3, input=inputs, options={"ncp": 3})
    results = pandas.DataFrame(
        results.data_matrix.T[1:, :],
        columns=["time"] + list(model.get_model_variables()),
    )

    for column in filter(lambda c: c != "time", results.columns):
        assert results[column].values == pytest.approx(data[column].values)


def test_FMU_integration_vector_problem(tmp_path, vector_problem):
    pyfmi = pytest.importorskip("pyfmi")

    # Setup the system
    solver = NonLinearSolver("solver", vector_problem)

    # Init
    vector_problem.in_ = 1.0
    vector_problem.add_driver(solver)
    vector_problem.run_drivers()

    # Convert to FMU
    fmu_file = to_fmu(
        vector_problem,
        dest=tmp_path,
        non_linear_solver=solver,
        inputs=["in_"],
        outputs=["out"],
        locals=["dummy_coef[0]"],
        fmu_name_suffix=str(uuid.uuid4()).replace("-", ""),
    )

    # Load the FMU
    model = pyfmi.load_fmu(str(fmu_file))

    # Run the FMU simulation
    inputs = ("in_", lambda t: 5.0 - 4.0 * numpy.exp(-t / 0.1))
    results = model.simulate(final_time=0.3, input=inputs, options={"ncp": 3})
    r = pandas.DataFrame(
        results.data_matrix.T[1:, :],
        columns=["time"] + list(model.get_model_variables()),
    )

    assert_almost_equal(r["dummy_coef[0]"].values, 0.5 * (r["in_"] + r["out"]))



import pytest

from cosapp.drivers import NonLinearMethods, NonLinearSolver
from cosapp.tests.library.systems import (
    SimpleTurbofan,
    AdvancedTurbofan,
    ComplexTurbofan,
)


def test_SimpleTurbofan_set_inputs():
    s = SimpleTurbofan("MySimpleTurbofan")

    s.exec_order = ["atm", "inlet", "fan", "duct", "noz"]
    s.duct.inwards.cst_loss = 0.98

    design = s.add_driver(NonLinearSolver("design"))
    design.runner.offdesign.add_unknown("fan.mech_in.PW")

    s.fan.mech_in.XN = 80
    # test inlet XN value
    assert s.fan.mech_in.XN == 80.0
    assert s.duct.inwards.cst_loss == 0.98


def test_SimpleTurbofan_running_80():
    s = SimpleTurbofan("MySimpleTurbofan")

    s.exec_order = ["atm", "inlet", "fan", "duct", "noz"]
    s.duct.inwards.cst_loss = 0.98

    design = s.add_driver(NonLinearSolver("design"))
    design.runner.offdesign.add_unknown("fan.mech_in.PW")

    s.fan.mech_in.XN = 80
    s.run_drivers()
    assert s.fan.mech_in.XN == 80.0
    assert s.fan.gh == pytest.approx(0.0385792701, rel=1e-5)
    assert s.fan.mech_in.PW == pytest.approx(8279856, rel=1e-5)
    # a second run changes fan power because of poor fsolve behavior
    # first run gives fan power residue bigger than 1e-5
    assert s.inlet.W_in.W == pytest.approx(161.463, rel=1e-4)


def test_SimpleTurbofan_running_100():
    s = SimpleTurbofan("MySimpleTurbofan")

    s.exec_order = ["atm", "inlet", "fan", "duct", "noz"]
    s.duct.inwards.cst_loss = 0.98

    design = s.add_driver(NonLinearSolver("design"))
    design.runner.offdesign.add_unknown("fan.mech_in.PW")

    s.fan.mech_in.XN = 100
    s.run_drivers()

    assert s.fan.mech_in.XN == 100.0
    assert s.fan.gh == pytest.approx(0.1566921207, rel=1e-5)
    assert s.fan.mech_in.PW == pytest.approx(10874232, rel=1e-5)
    assert s.inlet.W_in.W == pytest.approx(177.034, rel=1e-5)


def test_AdvancedTurbofan_set_inputs():
    s = AdvancedTurbofan("MyAdvancedTurbofan")

    s.exec_order = ["atm", "inlet", "fan", "merger", "duct", "bleed", "noz"]
    s.duct.inwards.cst_loss = 0.98

    design = s.add_driver(NonLinearSolver("design"))
    design.runner.offdesign.add_unknown("fan.mech_in.PW")

    s.fan.mech_in.XN = 80
    # test inlet XN value
    assert s.fan.mech_in.XN == 80.0
    assert s.duct.inwards.cst_loss == 0.98


def test_AdvancedTurbofan_running_80():
    s = AdvancedTurbofan("MyAdvancedTurbofan")

    s.exec_order = ["atm", "inlet", "fan", "merger", "duct", "bleed", "noz"]
    s.duct.inwards.cst_loss = 0.98

    design = s.add_driver(NonLinearSolver("design"))
    design.runner.offdesign.add_unknown("fan.mech_in.PW")

    s.fan.mech_in.XN = 80
    s.run_drivers()
    assert s.fan.mech_in.XN == 80.0
    assert s.fan.gh == pytest.approx(0.038579270081290681, rel=1e-5)
    assert s.fan.mech_in.PW == pytest.approx(8279856, rel=1e-5)
    assert s.inlet.W_in.W == pytest.approx(161.463, rel=1e-5)

    assert s.bleed.fl2_out.W == pytest.approx(1.630942952, rel=1e-5)
    assert s.merger.fl2_in.W == pytest.approx(1.630942952, rel=1e-5)
    assert s.bleed.fl2_out.Pt == pytest.approx(180023, rel=1e-5)
    assert s.merger.fl2_in.Pt == pytest.approx(180023, rel=1e-5)
    assert s.bleed.fl2_out.Tt == pytest.approx(324.2258, rel=1e-5)
    assert s.merger.fl2_in.Tt == pytest.approx(324.2258, rel=1e-5)


def test_AdvancedTurbofan_running_100():
    s = AdvancedTurbofan("MyAdvancedTurbofan")

    s.exec_order = ["atm", "inlet", "fan", "merger", "duct", "bleed", "noz"]
    s.duct.inwards.cst_loss = 0.98

    design = s.add_driver(NonLinearSolver("design"))
    design.runner.offdesign.add_unknown("fan.mech_in.PW")

    s.fan.mech_in.XN = 100
    s.run_drivers()

    assert s.fan.mech_in.XN == 100.0
    assert s.fan.gh == pytest.approx(0.15669212068943689, rel=1e-5)
    assert s.fan.mech_in.PW == pytest.approx(10874232, rel=1e-5)
    assert s.inlet.W_in.W == pytest.approx(177.034, rel=1e-5)

    assert s.bleed.fl2_out.W == pytest.approx(1.7882221046, rel=1e-5)
    assert s.merger.fl2_in.W == pytest.approx(1.7882221046, rel=1e-5)
    assert s.bleed.fl2_out.Pt == pytest.approx(200435.4, rel=1e-5)
    assert s.merger.fl2_in.Pt == pytest.approx(200435.4, rel=1e-5)
    assert s.bleed.fl2_out.Tt == pytest.approx(334.330, rel=1e-5)
    assert s.merger.fl2_in.Tt == pytest.approx(334.330, rel=1e-5)


def test_ComplexTurbofan_set_inputs():
    s = ComplexTurbofan("MyComplexTurbofan")
    s.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.bleed.inwards.split_ratio = 0.995
    s.bleed.inwards.split_ratio = 0.99

    design = s.add_driver(NonLinearSolver("design", tol=1e-7))
    design.runner.offdesign.add_unknown("fanC.mech_in.PW")

    s.fanC.mech_in.XN = 80
    # test inlet XN value
    assert s.fanC.mech_in.XN == 80.0
    assert s.duct.inwards.cst_loss == 0.98
    assert s.fanC.ductC.duct.inwards.cst_loss == 0.98


def test_ComplexTurbofan_running_80():
    s = ComplexTurbofan("MyComplexTurbofan")
    s.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.bleed.inwards.split_ratio = 0.995
    s.bleed.inwards.split_ratio = 0.99

    design = s.add_driver(NonLinearSolver("design", tol=1e-7))
    design.runner.offdesign.add_unknown("fanC.mech_in.PW")

    s.fanC.mech_in.XN = 80
    s.run_drivers()
    assert s.fanC.mech_in.XN == 80.0
    assert s.fanC.gh == pytest.approx(0.03857927008129068, rel=1e-5)
    assert s.fanC.mech_in.PW == pytest.approx(8114258.93, rel=1e-5)
    assert s.inlet.W_in.W == pytest.approx(158.2341, rel=1e-5)

    assert s.bleed.fl2_out.W == pytest.approx(1.59832409, rel=1e-5)
    assert s.merger.fl2_in.W == pytest.approx(1.59832409, rel=1e-5)
    assert s.bleed.fl2_out.Pt == pytest.approx(176422.55, rel=1e-5)
    assert s.merger.fl2_in.Pt == pytest.approx(176422.55, rel=1e-5)
    assert s.bleed.fl2_out.Tt == pytest.approx(324.2258, rel=1e-5)
    assert s.merger.fl2_in.Tt == pytest.approx(324.2258, rel=1e-5)


def test_ComplexTurbofan_internal_loop_ductC_running_80():
    # Another run with an internal loop at ductC level
    s = ComplexTurbofan("MyComplexTurbofan")
    s.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.bleed.inwards.split_ratio = 0.995
    s.bleed.inwards.split_ratio = 0.99

    design = s.add_driver(NonLinearSolver("design", tol=1e-7))
    design.runner.offdesign.add_unknown("fanC.mech_in.PW")

    s.fanC.mech_in.XN = 80
    s.fanC.ductC.add_driver(NonLinearSolver("design", method=NonLinearMethods.POWELL))
    s.run_drivers()

    assert s.fanC.mech_in.XN == 80.0
    assert s.fanC.gh == pytest.approx(0.03857927008, rel=1e-5)
    assert s.fanC.mech_in.PW == pytest.approx(8114258.932, rel=1e-5)
    assert s.inlet.W_in.W == pytest.approx(158.2341, rel=1e-5)

    assert s.bleed.fl2_out.W == pytest.approx(1.59832409, rel=1e-5)
    assert s.merger.fl2_in.W == pytest.approx(1.59832409, rel=1e-5)
    assert s.bleed.fl2_out.Pt == pytest.approx(176422.55, rel=1e-5)
    assert s.merger.fl2_in.Pt == pytest.approx(176422.55, rel=1e-5)
    assert s.bleed.fl2_out.Tt == pytest.approx(324.2258, rel=1e-5)
    assert s.merger.fl2_in.Tt == pytest.approx(324.2258, rel=1e-5)

    assert s.fanC.ductC.bleed.fl2_out.W == pytest.approx(0.79514614358, rel=1e-5)
    assert s.fanC.ductC.merger.fl2_in.W == pytest.approx(0.79514614358, rel=1e-5)
    assert s.fanC.ductC.bleed.fl2_out.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert s.fanC.ductC.merger.fl2_in.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert s.fanC.ductC.bleed.fl2_out.Tt == pytest.approx(273.15, rel=1e-5)
    assert s.fanC.ductC.merger.fl2_in.Tt == pytest.approx(273.15, rel=1e-5)


def test_ComplexTurbofan_internal_loop_fanC_running_80():
    # Another run with an internal loop at fanC level
    s = ComplexTurbofan("MyComplexTurbofan")
    s.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.bleed.inwards.split_ratio = 0.995
    s.bleed.inwards.split_ratio = 0.99

    s.add_driver(NonLinearSolver("design", tol=1e-7))

    s.fanC.mech_in.XN = 80
    design = s.fanC.add_driver(NonLinearSolver("design", tol=1e-7))
    design.runner.offdesign.add_unknown("mech_in.PW")
    s.run_drivers()

    assert s.fanC.mech_in.XN == 80.0
    assert s.fanC.gh / 0.038579270081290681 == pytest.approx(1, rel=1e-5)
    assert s.fanC.mech_in.PW / 8114258.9326799763 == pytest.approx(1.0, rel=1e-5)
    assert s.inlet.W_in.W / 158.23408418127087 == pytest.approx(1, rel=1e-5)

    assert s.bleed.fl2_out.W == pytest.approx(1.59832409, rel=1e-5)
    assert s.merger.fl2_in.W == pytest.approx(1.59832409, rel=1e-5)
    assert s.bleed.fl2_out.Pt == pytest.approx(176422.55, rel=1e-5)
    assert s.merger.fl2_in.Pt == pytest.approx(176422.55, rel=1e-5)
    assert s.bleed.fl2_out.Tt == pytest.approx(324.2258, rel=1e-5)
    assert s.merger.fl2_in.Tt == pytest.approx(324.2258, rel=1e-5)

    assert s.fanC.ductC.bleed.fl2_out.W == pytest.approx(0.79514614358, rel=1e-5)
    assert s.fanC.ductC.merger.fl2_in.W == pytest.approx(0.79514614358, rel=1e-5)
    assert s.fanC.ductC.bleed.fl2_out.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert s.fanC.ductC.merger.fl2_in.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert s.fanC.ductC.bleed.fl2_out.Tt == pytest.approx(273.15, rel=1e-5)
    assert s.fanC.ductC.merger.fl2_in.Tt == pytest.approx(273.15, rel=1e-5)


def test_ComplexTurbofan_running_100():
    s = ComplexTurbofan("MyComplexTurbofan")
    s.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.bleed.inwards.split_ratio = 0.995
    s.bleed.inwards.split_ratio = 0.99

    design = s.add_driver(NonLinearSolver("design", tol=1e-7))
    design.runner.offdesign.add_unknown("fanC.mech_in.PW")

    s.fanC.mech_in.XN = 100
    s.run_drivers()

    assert s.fanC.mech_in.XN == 100.0
    assert s.fanC.gh == pytest.approx(0.15669212069, rel=1e-5)
    assert s.fanC.mech_in.PW == pytest.approx(10656747.6, rel=1e-5)
    assert s.inlet.W_in.W == pytest.approx(173.4933, rel=1e-5)

    assert s.bleed.fl2_out.W == pytest.approx(1.75245766, rel=1e-5)
    assert s.merger.fl2_in.W == pytest.approx(1.75245766, rel=1e-5)
    assert s.bleed.fl2_out.Pt == pytest.approx(196426.7, rel=1e-5)
    assert s.merger.fl2_in.Pt == pytest.approx(196426.7, rel=1e-5)
    assert s.bleed.fl2_out.Tt == pytest.approx(334.330, rel=1e-5)
    assert s.merger.fl2_in.Tt == pytest.approx(334.330, rel=1e-5)


def to_ComplexTurbofan_html():
    s = ComplexTurbofan("MyComplexTurbofan")
    s.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.duct.inwards.cst_loss = 0.98
    s.fanC.ductC.bleed.inwards.split_ratio = 0.995
    s.bleed.inwards.split_ratio = 0.99

    s.add_driver(NonLinearSolver("design", tol=1e-7))

    s.to_html("complexturbofan.html")



import pytest

import numpy as np

from cosapp.core.numerics.boundary import AbstractTimeUnknown, TimeUnknown
from cosapp.core.eval_str import EvalString
from cosapp.systems import System
from cosapp.drivers.time.utils import TimeUnknownDict


# class DummyTimeUnknown(AbstractTimeUnknown):
#     def __init__(self, context, der, max_time_step=np.inf):
#         self.context = context
#         self.__der = EvalString(der, context)
#         self.__max_dt = EvalString(max_time_step, context)

#     @property
#     def der(self) -> EvalString:
#         """Expression of the time derivative, given as an EvalString"""
#         return self.__der

#     @property
#     def max_time_step_expr(self) -> EvalString:
#         """Expression of the maximum admissible time step, given as an EvalString."""
#         return self.__max_dt

#     def reset(self):
#         """Reset transient unknown to a reference value"""
#         pass


class SubSystem(System):
    def setup(self):
        self.add_inward('x', 1.0)
        self.add_inward('y', np.ones(3))


class DummySystem(System):
    def setup(self):
        self.add_child(SubSystem('sub1'))
        self.add_child(SubSystem('sub2'))

        self.add_inward('f', 0.0)


@pytest.fixture(scope="function")
def system():
    return DummySystem("system")


@pytest.fixture(scope="function")
def unknowns(system):
    """Returns a list of dummy unknowns"""
    a = TimeUnknown(system, "f", der=0.5)
    b = TimeUnknown(system, "sub1.y", der="sub1.x")
    c = TimeUnknown(system, "sub2.x", der="sub1.y**2", max_time_step="0.1 * sub1.x")
    return a, b, c


@pytest.fixture(scope="function")
def unknowns_dict(unknowns):
    a, b, c = unknowns
    return unknowns, TimeUnknownDict(A=a, B=b, C=c)


@pytest.mark.parametrize("mapping, expected", [
    (dict(), dict()),
    (dict(foo="bar"), dict(error=TypeError, match="invalid item 'foo': str")),
])
def test_TimeUnknownDict__init__(mapping, expected):
    error = expected.get('error', None)

    if error is None:
        vardict = TimeUnknownDict(**mapping)
        assert len(vardict) == expected.get('length', 0)
    else:
        pattern = expected.get('match', None)
        with pytest.raises(error, match=pattern):
            TimeUnknownDict(**mapping)


def test_TimeUnknownDict_getset(unknowns):
    """Test __getitem__ et __setitem__ methods"""
    a, b, c = unknowns

    vardict = TimeUnknownDict(A=a, B=b)
    assert set(vardict.keys()) == {"A", "B"}
    assert vardict["A"] is a
    assert vardict["B"] is b

    vardict["C"] = c
    assert set(vardict.keys()) == {"A", "B", "C"}
    assert vardict["C"] is c

    vardict["C2"] = c
    assert set(vardict.keys()) == {"A", "B", "C", "C2"}
    assert vardict["C2"] is c

    # Bad key in __getitem__
    with pytest.raises(KeyError):
        vardict["foo"]

    # Bad key in __setitem__
    with pytest.raises(TypeError, match="invalid key 0"):
        vardict[0] = a

    # Bad value in __setitem__
    with pytest.raises(TypeError, match="invalid item 'foo': str"):
        vardict["foo"] = "bar"

    with pytest.raises(TypeError, match="invalid item 'foo': int"):
        vardict["foo"] = 0


def test_TimeUnknownDict__len__(unknowns):
    a, b, c = unknowns

    vardict = TimeUnknownDict(A=a, B=b)
    assert len(vardict) == 2
    vardict["C"] = c
    assert len(vardict) == 3

    assert len(TimeUnknownDict()) == 0
    assert len(TimeUnknownDict(A=a)) == 1
    assert len(TimeUnknownDict(A=a, B=b)) == 2
    assert len(TimeUnknownDict(A=a, B=b, C=c)) == 3


def test_TimeUnknownDict_update(system, unknowns):
    a, b, c = unknowns

    vardict = TimeUnknownDict(A=a)
    assert set(vardict.keys()) == {"A"}

    vardict.update({"B": b, "C": c})
    assert set(vardict.keys()) == {"A", "B", "C"}

    d = TimeUnknown(system, 'sub2.y', der='f', max_time_step=0.1)
    other = TimeUnknownDict(D=d)
    vardict.update(other)
    assert vardict["D"] is d
    assert set(vardict.keys()) == {"A", "B", "C", "D"}
    assert set(vardict.values()) == {a, b, c, d}


def test_TimeUnknownDict_pop(unknowns):
    a, b, c = unknowns

    vardict = TimeUnknownDict(A=a, B=b, C=c)
    assert set(vardict.keys()) == {"A", "B", "C"}

    assert vardict.pop("B") is b
    assert set(vardict.keys()) == {"A", "C"}

    assert vardict.pop("foo", None) is None
    assert vardict.pop("foo", 3.14) == 3.14

    with pytest.raises(KeyError):
        vardict.pop("foo")


def test_TimeUnknownDict__contains__(unknowns_dict):
    unknowns, vardict = unknowns_dict
    assert set(vardict.keys()) == {"A", "B", "C"}
    assert "A" in vardict
    assert "B" in vardict
    assert "C" in vardict
    assert "D" not in vardict

    vardict.pop("B")
    assert "A" in vardict
    assert "B" not in vardict
    assert "C" in vardict


def test_TimeUnknownDict_clear(unknowns_dict):
    vardict = unknowns_dict[1]
    assert set(vardict.keys()) == {"A", "B", "C"}
    
    vardict.clear()
    assert set(vardict.keys()) == set()


def test_TimeUnknownDict_constrained(unknowns_dict):
    """Test property `constrained`"""
    (a, b, c), vardict = unknowns_dict
    constrained = vardict.constrained
    assert isinstance(constrained, dict)
    assert len(constrained) == 1
    assert set(constrained.keys()) <= set(vardict.keys())
    assert set(constrained.keys()) == {"C"}

    for key, value in constrained.items():
        assert value.constrained

    # Check that property is a shallow copy of an internal dict
    assert vardict.constrained.get("C") is c
    assert vardict.constrained.pop("C") is c
    assert set(vardict.constrained.keys()) == {"C"}  # key still present after pop


def test_TimeUnknownDict_iterators(unknowns_dict, system):
    """Test keys() and values() methods"""
    (a, b, c), vardict = unknowns_dict

    keys, values = [], []
    for key, value in vardict.items():
        assert isinstance(key, str)
        assert isinstance(value, AbstractTimeUnknown)
        keys.append(key)
        values.append(value)

    assert len(vardict) == 3
    assert len(keys) == len(vardict)
    assert len(values) == len(vardict)
    assert set(keys) == {"A", "B", "C"}
    assert set(values) == {a, b, c}

    assert set(vardict.keys()) == {"A", "B", "C"}
    assert set(vardict.values()) == {a, b, c}
    assert set(vardict.keys(constrained=True)) == {"C"}
    assert set(vardict.values(constrained=True)) == {c}

    vardict["C2"] = vardict["C"]
    assert vardict["C2"] is c
    assert set(vardict.keys()) == {"A", "B", "C", "C2"}
    assert set(vardict.values()) == {a, b, c}
    assert set(vardict.keys(constrained=True)) == {"C", "C2"}
    assert set(vardict.values(constrained=True)) == {c}

    vardict["C2"] = vardict["A"]
    assert vardict["C2"] is a
    assert set(vardict.keys()) == {"A", "B", "C", "C2"}
    assert set(vardict.values()) == {a, b, c}
    assert set(vardict.keys(constrained=True)) == {"C"}
    assert set(vardict.values(constrained=True)) == {c}

    # Add new time unknown
    d = TimeUnknown(system, 'sub2.y', der='f', max_time_step=0.1)
    vardict["D"] = d
    assert set(vardict.keys()) == {"A", "B", "C", "C2", "D"}
    assert set(vardict.values()) == {a, b, c, d}
    assert set(vardict.keys(constrained=True)) == {"C", "D"}
    assert set(vardict.values(constrained=True)) == {c, d}
    # Check property `constrained`
    assert isinstance(vardict.constrained, dict)
    assert set(vardict.keys(constrained=True)) == set(vardict.constrained.keys())
    assert set(vardict.values(constrained=True)) == set(vardict.constrained.values())


def test_TimeUnknownDict_max_time_step(unknowns_dict):
    unknowns, vardict = unknowns_dict
    assert set(vardict.keys()) == {"A", "B", "C"}
    
    s = vardict["C"].context
    s.sub1.x = 3
    s.sub1.y = np.r_[1.2, 0]
    s.sub2.x = 0
    s.sub2.y[:] = 0

    dt = vardict.max_time_step()
    assert dt == min(var.max_time_step for var in unknowns)
    assert dt == pytest.approx(0.3, rel=1e-15)

    s.sub1.x = -0.5
    with pytest.raises(RuntimeError, match="maximum time step of C was evaluated to non-positive value -0.05"):
        vardict.max_time_step()
    


from typing import Any


class ContextLock:
    """Simple locker to track an execution task is going on."""

    def __init__(self):
        self.__locked = False

    def __enter__(self):
        self.__locked = True

    def __exit__(self, exc_type, exc_value, traceback):
        self.__locked = False

    def is_locked(self) -> bool:
        """Is the context lock locked?"""
        return self.__locked



"""
Various small helper functions.
"""
import enum
import inspect
import os
from numbers import Number
from typing import Any, Callable, Iterable, Type, Union, Tuple
from collections.abc import Collection

import numpy


def is_numerical(value: Any) -> bool:
    """Test if a value is numerical based on its type.

    Currently type considered as numerical are:

    - int; but not bool
    - float; including numpy.inf and numpy.nan
    - complex
    - numpy.ndarray of dtype numpy.number
    - Collection convertible in a numpy array of dtype derived from numpy.number

    Parameters
    ----------
    value : Any
        Value to test

    Returns
    -------
    bool
        Is the value numerical type?
    """
    if isinstance(value, (bool, str)):
        # to avoid bool being considered as int
        # and because str is Collection
        return False
    if isinstance(value, Number):
        return True
    if isinstance(value, numpy.ndarray):
        return numpy.issubdtype(value.dtype, numpy.number)
    if isinstance(value, Collection) and len(value) > 0:
        return all(is_numerical(v) for v in value)
    return False


def is_number(value: Any) -> bool:
    """Test if a value is a Number or a 0d numerical array

    Currently type considered as numerical are:

    - int; but not bool
    - float; including numpy.inf and numpy.nan
    - complex
    - numpy.ndarray of dtype numpy.number and ndim == 0

    Parameters
    ----------
    value : Any
        Value to test

    Returns
    -------
    bool
        Is the value numerical type?
    """
    if isinstance(value, bool):
        # to avoid bool being considered as int
        return False
    if isinstance(value, Number):
        return True
    if isinstance(value, numpy.ndarray) and value.ndim == 0:
        return numpy.issubdtype(value.dtype, numpy.number)
    return False


def get_typename(dtype: Union[Type, Tuple[Type]], multiformat="({})") -> str:
        if inspect.isclass(dtype):
            return dtype.__qualname__
        return multiformat.format(", ".join(set(t.__qualname__ for t in dtype)))


def check_arg(
    arg: Any,
    argname: str,
    dtype: Union[Type, Iterable[Type]],
    value_ok: Callable[[Any], bool] = None
):
    """
    Utility function for argument type and value validation.
    Raises a TypeError exception if type(arg) is not in type list given by 'dtype'.
    Raises a ValueError exception if value_ok(arg) is False, where 'value_ok' is a
    boolean function defining a validity criterion.

    For example:
    >>> check_arg(-0.12, 'my_var', float)
    does not raise any exception, as -0.12 is a float

    >>> check_arg(-0.12, 'my_var', (int, str))
    raises TypeError, as first argument is neither an int, not a str

    >>> check_arg(-0.12, 'my_var', float, value_ok = lambda x: x > 0)
    raises ValueError, as first argument is not strictly positive
    """

    def get_caller():
        stack = inspect.stack()
        return stack[3] if len(stack) > 3 else stack[-1]

    def get_context(caller):
        try:
            context = caller.code_context[0]
        except:
            context = ""
        return os.path.basename(caller.filename), caller.lineno, context

    # Check type
    if not isinstance(arg, dtype):
        valid = get_typename(dtype, multiformat="one of ({})")
        caller = get_caller()
        raise TypeError("argument '{}' should be {}; got {} {!r}\nIn {}, line #{}: \n{}".format(
            argname, valid, type(arg).__qualname__, arg, *get_context(caller)))
    # Check value
    if isinstance(value_ok, Callable) and not value_ok(arg):
        caller = get_caller()
        raise ValueError("argument {!r} was given invalid value {!r}\nIn {}, line #{}: \n{}".format(
            argname, arg, *get_context(caller)))



# -*- coding: utf-8 -*-
"""
    sphinx-mermaid
    ~~~~~~~~~~~~~~~

    https://github.com/mgaitan/sphinxcontrib-mermaid
    Modified for the purpose of CoSApp

    Allow mermaid diagramas to be included in Sphinx-generated
    documents inline.

    :copyright: Copyright 2016 by Martín Gaitán and others, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import codecs
import posixpath
import os
from subprocess import Popen, PIPE
from hashlib import sha1
from tempfile import _get_default_tempdir
from six import text_type
from docutils import nodes
from docutils.parsers.rst import Directive, directives
from docutils.statemachine import ViewList

import sphinx
from sphinx.application import Sphinx
from sphinx.errors import SphinxError
from sphinx.locale import _
from sphinx.util.i18n import search_image_for_language
from sphinx.util.fileutil import copy_asset
from sphinx.util.osutil import ensuredir, ENOENT
from sphinx.util import logging

logger = logging.getLogger(__name__)


class MermaidError(SphinxError):
    category = "Mermaid error"


class mermaid(nodes.General, nodes.Inline, nodes.Element):
    pass


def figure_wrapper(directive, node, caption):
    figure_node = nodes.figure("", node)
    if "align" in node:
        figure_node["align"] = node.attributes.pop("align")

    parsed = nodes.Element()
    directive.state.nested_parse(
        ViewList([caption], source=""), directive.content_offset, parsed
    )
    caption_node = nodes.caption(parsed[0].rawsource, "", *parsed[0].children)
    caption_node.source = parsed[0].source
    caption_node.line = parsed[0].line
    figure_node += caption_node
    return figure_node


def align_spec(argument):
    return directives.choice(argument, ("left", "center", "right"))


class Mermaid(Directive):
    """
    Directive to insert arbitrary Mermaid markup.
    """

    has_content = True
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = False
    option_spec = {
        "alt": directives.unchanged,
        "align": align_spec,
        "caption": directives.unchanged,
    }

    def get_mm_code(self):
        if self.arguments:
            # try to load mermaid code from an external file
            document = self.state.document
            if self.content:
                return [
                    document.reporter.warning(
                        "Mermaid directive cannot have both content and "
                        "a filename argument",
                        line=self.lineno,
                    )
                ]
            env = self.state.document.settings.env
            argument = search_image_for_language(self.arguments[0], env)
            rel_filename, filename = env.relfn2path(argument)
            env.note_dependency(rel_filename)
            try:
                with codecs.open(filename, "r", "utf-8") as fp:
                    mmcode = fp.read()
            except (IOError, OSError):
                return [
                    document.reporter.warning(
                        "External Mermaid file %r not found or reading "
                        "it failed" % filename,
                        line=self.lineno,
                    )
                ]
        else:
            # inline mermaid code
            mmcode = "\n".join(self.content)
            if not mmcode.strip():
                return [
                    self.state_machine.reporter.warning(
                        'Ignoring "mermaid" directive without content.',
                        line=self.lineno,
                    )
                ]
        return mmcode

    def run(self):

        node = mermaid()
        node["code"] = self.get_mm_code()
        node["options"] = {}
        if "alt" in self.options:
            node["alt"] = self.options["alt"]
        if "align" in self.options:
            node["align"] = self.options["align"]
        if "inline" in self.options:
            node["inline"] = True

        caption = self.options.get("caption")
        if caption:
            node = figure_wrapper(self, node, caption)

        return [node]


def render_mm(self, code, options, format, prefix="mermaid"):
    """Render mermaid code into a PNG or PDF output file."""

    if format == "raw":
        format = "png"

    mermaid_cmd = self.builder.config.mermaid_cmd
    hashkey = (
        code + str(options) + str(self.builder.config.mermaid_sequence_config)
    ).encode("utf-8")

    basename = "%s-%s" % (prefix, sha1(hashkey).hexdigest())
    fname = "%s.%s" % (basename, format)
    relfn = posixpath.join(self.builder.imgpath, fname)
    outdir = os.path.join(self.builder.outdir, self.builder.imagedir)
    outfn = os.path.join(outdir, fname)
    tmpfn = os.path.join(_get_default_tempdir(), basename)

    if os.path.isfile(outfn):
        return relfn, outfn

    ensuredir(os.path.dirname(outfn))

    # mermaid expects UTF-8 by default
    if isinstance(code, text_type):
        code = code.encode("utf-8")

    with open(tmpfn, "wb") as t:
        t.write(code)

    mm_args = [mermaid_cmd, "-i", tmpfn, "-o", outfn]
    mm_args.extend(self.builder.config.mermaid_params)
    if self.builder.config.mermaid_sequence_config:
        mm_args.extend("--configFile", self.builder.config.mermaid_sequence_config)

    try:
        p = Popen(mm_args, stdout=PIPE, stdin=PIPE, stderr=PIPE)
    except OSError as err:
        if err.errno != ENOENT:  # No such file or directory
            raise
        logger.warning(
            "command %r cannot be run (needed for mermaid "
            "output), check the mermaid_cmd setting" % mermaid_cmd
        )
        return None, None

    stdout, stderr = p.communicate(code)
    if self.builder.config.mermaid_verbose:
        logger.info(stdout)

    if p.returncode != 0:
        raise MermaidError(
            "Mermaid exited with error:\n[stderr]\n%s\n"
            "[stdout]\n%s" % (stderr, stdout)
        )
    if not os.path.isfile(outfn):
        raise MermaidError(
            "Mermaid did not produce an output file:\n[stderr]\n%s\n"
            "[stdout]\n%s" % (stderr, stdout)
        )
    return relfn, outfn


def _render_mm_html_raw(
    self, node, code, options, prefix="mermaid", imgcls=None, alt=None
):
    if "align" in node:
        tag_template = """<div align="{align}" class="mermaid align-{align}">
            {code}
        </div>
        """
    else:
        tag_template = """<div class="mermaid">
            {code}
        </div>"""

    self.body.append(
        tag_template.format(align=node.get("align"), code=self.encode(code))
    )
    raise nodes.SkipNode


def render_mm_html(self, node, code, options, prefix="mermaid", imgcls=None, alt=None):
    format = self.builder.config.mermaid_output_format
    if format == "raw":
        return _render_mm_html_raw(
            self, node, code, options, prefix="mermaid", imgcls=None, alt=None
        )

    try:
        if format not in ("png", "svg"):
            raise MermaidError(
                "mermaid_output_format must be one of 'raw', 'png', "
                "'svg', but is %r" % format
            )

        fname, outfn = render_mm(self, code, options, format, prefix)
    except MermaidError as exc:
        logger.warning("mermaid code %r: " % code + str(exc))
        raise nodes.SkipNode

    if fname is None:
        self.body.append(self.encode(code))
    else:
        if alt is None:
            alt = node.get("alt", self.encode(code).strip())
        imgcss = imgcls and 'class="%s"' % imgcls or ""
        if format == "svg":
            svgtag = """<object data="%s" type="image/svg+xml">
            <p class="warning">%s</p></object>\n""" % (
                fname,
                alt,
            )
            self.body.append(svgtag)
        else:
            if "align" in node:
                self.body.append(
                    '<div align="%s" class="align-%s">' % (node["align"], node["align"])
                )

            self.body.append('<img src="%s" alt="%s" %s/>\n' % (fname, alt, imgcss))
            if "align" in node:
                self.body.append("</div>\n")

    raise nodes.SkipNode


def html_visit_mermaid(self, node):
    render_mm_html(self, node, node["code"], node["options"])


def render_mm_latex(self, node, code, options, prefix="mermaid"):
    try:
        fname, outfn = render_mm(self, code, options, "pdf", prefix)
    except MermaidError as exc:
        logger.warning("mm code %r: " % code + str(exc))
        raise nodes.SkipNode

    if self.builder.config.mermaid_pdfcrop != "":
        mm_args = [self.builder.config.mermaid_pdfcrop, outfn]
        try:
            p = Popen(mm_args, stdout=PIPE, stdin=PIPE, stderr=PIPE)
        except OSError as err:
            if err.errno != ENOENT:  # No such file or directory
                raise
            logger.warning(
                "command %r cannot be run (needed to crop pdf), check the mermaid_cmd setting"
                % self.builder.config.mermaid_pdfcrop
            )
            return None, None

        stdout, stderr = p.communicate()
        if self.builder.config.mermaid_verbose:
            logger.info(stdout)

        if p.returncode != 0:
            raise MermaidError(
                "PdfCrop exited with error:\n[stderr]\n%s\n"
                "[stdout]\n%s" % (stderr, stdout)
            )
        if not os.path.isfile(outfn):
            raise MermaidError(
                "PdfCrop did not produce an output file:\n[stderr]\n%s\n"
                "[stdout]\n%s" % (stderr, stdout)
            )

        fname = "{filename[0]}-crop{filename[1]}".format(
            filename=os.path.splitext(fname)
        )

    is_inline = self.is_inline(node)
    if is_inline:
        para_separator = ""
    else:
        para_separator = "\n"

    if fname is not None:
        post = None
        if not is_inline and "align" in node:
            if node["align"] == "left":
                self.body.append("{")
                post = "\\hspace*{\\fill}}"
            elif node["align"] == "right":
                self.body.append("{\\hspace*{\\fill}")
                post = "}"
        self.body.append(
            "%s\\sphinxincludegraphics{%s}%s" % (para_separator, fname, para_separator)
        )
        if post:
            self.body.append(post)

    raise nodes.SkipNode


def latex_visit_mermaid(self, node):
    render_mm_latex(self, node, node["code"], node["options"])


def render_mm_texinfo(self, node, code, options, prefix="mermaid"):
    try:
        fname, outfn = render_mm(self, code, options, "png", prefix)
    except MermaidError as exc:
        logger.warning("mm code %r: " % code + str(exc))
        raise nodes.SkipNode
    if fname is not None:
        self.body.append("@image{%s,,,[mermaid],png}\n" % fname[:-4])
    raise nodes.SkipNode


def texinfo_visit_mermaid(self, node):
    render_mm_texinfo(self, node, node["code"], node["options"])


def text_visit_mermaid(self, node):
    if "alt" in node.attributes:
        self.add_text(_("[graph: %s]") % node["alt"])
    else:
        self.add_text(_("[graph]"))
    raise nodes.SkipNode


def man_visit_mermaid(self, node):
    if "alt" in node.attributes:
        self.body.append(_("[graph: %s]") % node["alt"])
    else:
        self.body.append(_("[graph]"))
    raise nodes.SkipNode


def config_inited(app, config):
    version = config.mermaid_version
    mermaid_js_url = "https://unpkg.com/mermaid@{}/dist/mermaid.min.js".format(version)
    app.add_js_file(mermaid_js_url)
    app.add_js_file(
        None, 
        body='mermaid.initialize({startOnLoad:true, theme:"neutral", sequenceConfig: {mirrorActors: false}});'
    )
    app.add_css_file("mermaid.css")


def on_build_finished(app: Sphinx, exc: Exception) -> None:
    if exc is None:
        src = os.path.join(os.path.dirname(__file__), "mermaid.css")
        dst = os.path.join(app.outdir, "_static")
        copy_asset(src, dst)


def setup(app):
    app.add_node(
        mermaid,
        html=(html_visit_mermaid, None),
        latex=(latex_visit_mermaid, None),
        texinfo=(texinfo_visit_mermaid, None),
        text=(text_visit_mermaid, None),
        man=(man_visit_mermaid, None),
    )
    app.add_directive("mermaid", Mermaid)

    #
    app.add_config_value("mermaid_cmd", "mmdc", "html")
    app.add_config_value("mermaid_pdfcrop", "", "html")
    app.add_config_value("mermaid_output_format", "raw", "html")
    app.add_config_value("mermaid_params", list(), "html")
    app.add_config_value("mermaid_verbose", False, "html")
    app.add_config_value("mermaid_sequence_config", False, "html")
    app.add_config_value("mermaid_version", "8.5.0", "html")

    app.connect("config-inited", config_inited)
    app.connect('build-finished', on_build_finished)

    return {"version": sphinx.__display_version__, "parallel_read_safe": True}



from collections import OrderedDict
from typing import Any, Dict, List, Set, Type, Union

from cosapp.recorders import DataFrameRecorder
from cosapp.systems import MetaSystem
from cosapp.systems.surrogate_models.response_surface import ResponseSurface
from cosapp.drivers.driver import Driver
from cosapp.drivers.lineardoe import LinearDoE
from cosapp.drivers.nonlinearsolver import NonLinearSolver


# TODO
# [ ] Quid for vector variables
# [ ] Simplify the builder by using inputs and outputs of the targeted system.
class MetaSystemBuilder(Driver):

    __slots__ = ('responses', 'model_type', '_metasystem')

    def __init__(
        self, name: str, owner: "Optional[cosapp.systems.System]" = None, **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)

        self.responses = set()  # type: Set[str]
        self.model_type = ResponseSurface  # type: Type
        self._metasystem = None  # type: cosapp.systems.MetaSystem

        self.add_child(LinearDoE("doe"))
        self.doe.add_child(NonLinearSolver("solve"))

    def compute(self):
        includes = list(self.doe.input_vars)
        includes.extend(self.responses)
        self.doe.add_recorder(DataFrameRecorder(raw_output=True, includes=includes))
        self.doe.run_once()

        class MyMeta(MetaSystem):
            def setup(self, model_type=ResponseSurface, ref_model=None):
                self._default_model = model_type

                if ref_model:
                    self.inputs = OrderedDict(ref_model.inputs)
                    self.outputs = OrderedDict(ref_model.outputs)

        self._metasystem = MyMeta(
            "meta",
            self.doe.recorder.export_data(),
            model_type=self.model_type,
            ref_model=self.owner,
        )

    def add_input_var(
        self,
        definition: Union[str, Dict[str, Any]],
        lower: float = None,
        upper: float = None,
        count: int = 2,
    ):
        self.doe.add_input_var(definition, lower, upper, count)

    def add_response(self, name: Union[str, List[str]]):

        if not isinstance(name, (str, list)):
            raise TypeError(
                f"'name' should be a string or a list of strings; got {type(name).__name__}."
            )

        def add_unique_response_var(name: str):
            self.check_owner_attr(name)
            self.responses.add(name)

        if isinstance(name, str):
            add_unique_response_var(name)
        else:
            for n in name:
                if isinstance(n, str):
                    add_unique_response_var(n)
                else:
                    raise TypeError(
                        f"name list should only contain strings; got {type(n).__name__}."
                    )



class Singleton(type):
    """
    Define an instance operation that lets clients access its unique instance.

    Examples
    --------

    >>> class MyClass(metaclass=Singleton):
    >>>     pass
    >>> 
    >>> m1 = MyClass()
    >>> m2 = MyClass()
    >>> assert m1 is m2
    """
    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs)
        cls.__instance = None

    def __call__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__call__(*args, **kwargs)
        return cls.__instance



"""Build a D3js view of a System."""
import os

from cosapp.systems import System
from cosapp.tools import templates


def to_d3(system: System, show: bool = True, size: int = 300):
    """Returns the representation of this system in HTML format.

    Returns
    -------
    str
        HTML formatted representation
    """
    # Optional dependencies
    try:
        from jinja2 import Environment, PackageLoader
    except ImportError:
        raise ImportError("jinja2 needs to be installed to export to d3 JS.")

    def build_d3_repr(
        system: System, filename: str, embeddable: bool = False, size: int = 600
    ) -> None:
        env = Environment(
            loader=PackageLoader("cosapp.tools", "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = env.get_template("system_d3.html")

        # This is inspired from openmdao problem_viewer
        html_begin_tags = """<!DOCTYPE html>
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
<title>{{title}}</title>
</head>
<body>
<div class="flexdiv">
"""
        html_end_tags = """
</div>
</body>
</html>
"""

        ressource_dir = os.path.dirname(os.path.abspath(templates.__file__))
        libs_dir = os.path.join(ressource_dir, "lib")
        src_dir = os.path.join(ressource_dir, "src")

        # Grab the resources
        with open(os.path.join(libs_dir, "d3.min.js"), "r", encoding="utf-8") as f:
            d3_js = f.read()
        with open(os.path.join(src_dir, "d3_draw.js"), "r", encoding="utf-8") as f:
            draw_js = f.read()
        with open(os.path.join(src_dir, "d3_styles.css"), "r", encoding="utf-8") as f:
            d3_styles = f.read()

        def build_d3_json(
            system: System, parent_name: str = None, nlevels: int = 0
        ) -> dict:

            if parent_name is None:
                full_name = system.name
            else:
                full_name = f"{parent_name}.{system.name}"

            if nlevels == 0:  # Number of System levels
                def get_level(syst) -> int:
                    if len(syst.children) == 0:
                        return 1
                    else:
                        return 1 + max(map(get_level, syst.children.values()))

                nlevels = get_level(system)

            level = len(full_name.split("."))

            rtn = {"name": system.name, "full_name": full_name}
            if system.children:
                rtn["children"] = [
                    build_d3_json(system[child], parent_name=full_name, nlevels=nlevels)
                    for child in system.exec_order
                ]
            else:
                rtn["size"] = (nlevels / level) ** 5

            return rtn

        elements = {
            "title": "my viewer",
            "model_data": f"var modelData = {build_d3_json(system)!s}",
            "d3_styles": d3_styles,
            "d3JS": d3_js,
            "draw": draw_js,
        }

        rendered_html = template.render(elements)
        if not embeddable:
            html_begin_tags = html_begin_tags.replace("{{title}}", elements["title"])
            rendered_html = html_begin_tags + rendered_html + html_end_tags

        _, ext = os.path.splitext(filename)
        if ext != ".html":
            filename = f"{filename}{os.extsep}html"

        with open(filename, mode="w", encoding="utf-8") as f:
            f.write(rendered_html)

    build_d3_repr(system, f"{system.name}.html")

    if show:
        # Optional dependencies
        try:
            from IPython.display import IFrame
        except ImportError:
            raise ImportError("IPython needs to be installed to display the d3 figure.")
        return IFrame(temp_name, "99.9%", "435px")
    else:
        return None




"""Class defining a triangular distribution."""
import numbers
from typing import Any, Dict, Optional

import numpy
import scipy.stats
import scipy.optimize

from .distribution import Distribution


class Triangular(Distribution):
    """A class to define a triangular distribution.

    Parameters
    ----------
    worst : float
        The parameter absolute worst value.
    likely : float
        The parameter absolute likely value (i.e. corresponding to the peak probability).
    best : float
        The parameter absolute best value.
    pworst : float, optional
        The worst value probability is the probability that the variable will be lower (if worst<best) or higher
        (if worst>best) than the worst value; default 0.15 (i.e. 15%).
    pbest : float, optional
        The best value probability is the probability that the variable will be higher (if worst<best) or lower
        (if worst>best) than the best value; default 0.15 (i.e. 15%).
    """

    def __init__(
        self,
        worst: float,
        likely: float,
        best: float,
        pworst: Optional[float] = 0.15,
        pbest: Optional[float] = 0.15,
    ):
        self._rv = None  # type: scipy.stats.triang
        # Dummy init
        self._likely = 0.5 * (best + worst)  # type: numbers.Number
        super().__init__(worst, best, pworst, pbest)

        self.likely = likely  # Trigger likely validation

    def __json__(self) -> Dict[str, Any]:
        """Serialize the distribution object.

        Returns
        -------
        Dict[str, Any]
            JSONable dictionary describing the distribution.
        """
        base = super().__json__()
        base.update({"likely": self.likely})
        return base

    @property
    def likely(self) -> float:
        """float : The parameter absolute likely value
        
        It corresponds to the peak probability.
        """
        return self._likely

    @likely.setter
    def likely(self, value: float):
        params = self._rv.kwds
        lower = params["loc"]
        upper = lower + params["scale"]
        if not (lower <= value <= upper):
            raise ValueError(
                f"Likely value not within distribution bounds: {lower} <= {value} <= {upper}."
            )
        self._likely = value
        self._set_distribution()

    def _set_distribution(self) -> None:
        """Set the probability distribution according the parameters."""
        if self.pworst + self.pbest > 1.0:
            raise ValueError(
                f"Best and worst probabilities are incompatible: {self.__json__()!s}."
            )

        pts = [self.worst, self.best]
        if self.worst > self.best:
            ppts = [1 - self.pworst, self.pbest]
        else:
            ppts = [self.pworst, 1 - self.pbest]

        if self._rv is None:
            x0 = [min(0, self.likely), 2 * abs(self.likely)]
        else:
            params = self._rv.kwds  # return {"c": #, "loc": #, "scale": #}
            x0 = [params["loc"], params["scale"]]

        def make_triang(x):
            # likely = loc + c * scale
            if any(numpy.isnan(x)):
                raise ValueError(f"invalid distribution parameters {x}")
            c = (self.likely - x[0]) / x[1]
            return scipy.stats.triang(c=c, loc=x[0], scale=x[1])

        def f(x):
            t = make_triang(x)
            return t.ppf(ppts) - pts

        res = scipy.optimize.root(f, x0)
        if not res.success or any(numpy.isnan(res.x)):
            raise ValueError(
                f"Unable to fit triangular distribution on {self.__json__()!s}."
            )
        self._rv = make_triang(res.x)

    def draw(self, quantile: Optional[float] = None) -> float:
        """Generate a random number.

        If a quantile is given, generate the perturbation for that quantile.

        Parameters
        ----------
        quantile : Optional[float], optional
            Quantile for which the perturbation must be set; default None (i.e. random perturbation)
        
        Returns
        -------
        float
            The random number
        """
        return self._rv.rvs() if quantile is None else self._rv.ppf(quantile)



import pytest

import numpy as np

from cosapp.systems import System
from cosapp.drivers.time.utils import TimeVarManager
from cosapp.core.numerics.basics import MathematicalProblem
from .conftest import (
    ScalarOde, VectorOde,
    PointMass, PointMassWithPorts,
    ExpRampOde, DampedMassSpring,
    CoupledTanks,
    )


@pytest.mark.parametrize("ders, expected", [
    ({}, {}),
    ({'h': 'Q / A'}, {'h': ['h', 'Q / A']}),
    ({'x': 'v'}, {'x': ['x', 'v']}),
    ({'v': 'a', 'x': 'v'}, {'x': ['x', 'v', 'a']}),
    ({'v': 'a', 'x': 'v', 'f': 'dfdt'}, {'x': ['x', 'v', 'a'], 'f': ['f', 'dfdt']}),
    ({'v': 'a', 'x': 'v', 'y': 'v', 'h': 'y'}, {'x': ['x', 'v', 'a'], 'h': ['h', 'y', 'v', 'a']}),
    ({'x': 'v', 'v': 'x'}, {}),
])
def test_TimeVarManager_get_tree(ders, expected):
    tree = TimeVarManager.get_tree(ders)
    assert tree == expected


@pytest.mark.parametrize("system, expected", [
    (PointMass('point'), dict(transients=["[x, v]"])),
    (PointMassWithPorts('point'), dict(transients=["[position.x, kinematics.v]"])),
])
def test_TimeVarManager__init__(system, expected):
    m = TimeVarManager(system)
    assert m.context is system
    assert isinstance(m.problem, MathematicalProblem)
    assert m.problem.context is system
    assert set(m.transients.keys()) == set(expected.get("transients", []))
    assert set(m.rates.keys()) == set(expected.get("rates", []))


def test_TimeVarManager_context():
    a = PointMass("A")
    b = PointMass("B")
    m = TimeVarManager(a)
    assert m.context is a
    m.context = b
    assert m.context is b
    m.context = a
    assert m.context is a

    for value in [1, "A"]:
        with pytest.raises(TypeError):
            m.context = value


def test_TimeVarManager_max_time_step(clock):
    class DynamicSystem(System):
        def setup(self):
            self.add_inward('a', np.ones(3))
            self.add_inward('slope', 1.5)
            self.add_inward('omega', np.pi)

            self.add_transient('v', der='a')
            self.add_transient('x', der='v', max_time_step='0.2 * norm(v) / norm(a)')
            self.add_transient('h', der='slope', max_time_step='cos(omega * t)')

    system = DynamicSystem('system')
    m = TimeVarManager(system)

    clock.reset()
    system.a = np.ones(3)
    system.v = np.ones(3)
    assert m.max_time_step() == pytest.approx(0.2)  # limited by transient 'x'
    system.a = np.ones(3)
    system.v = np.zeros(3)
    with pytest.raises(RuntimeError,
        match=r"The maximum time step of \[x, v\] was evaluated to non-positive value 0"
    ):
        m.max_time_step()

    system.a = np.r_[0, 0, 1]
    system.v = np.r_[100, 0, 0]
    assert m.max_time_step() == pytest.approx(1)  # limited by transient 'h'

    clock.time = 1.2
    with pytest.raises(RuntimeError,
        match=r"The maximum time step of h was evaluated to non-positive value -0..."
    ):
        m.max_time_step()



from .multiply import (
    MultiplyFamily,
    Multiply1,
    Multiply2,
    Multiply2Derivative,
    Multiply3,
    Multiply4,
    MultiplySystem,
    MultiplySystem2,
    MultiplyVector2,
    MultiplyVector3,
    NonLinear1,
    NonLinear3,
    IterativeNonLinear,
    Multiply2Derivative,
    IterativeNonLinearDerivative,
)
from .multiply import Splitter as SplitterMath
from .multiply import Merger as MergerMath
from .others import (
    ComplexTurbofan,
    SimpleTurbofan,
    AdvancedTurbofan,
    ComplexDuct,
    Duct,
    Fan,
    SerialDuct,
    RealDuct,
    TurbofanFamily,
    PlossFamily,
    Splitter,
    Merger,
    Atm,
    Inlet,
    Nozzle,
    FanComplex,
)
from .pressurelossvarious import (
    PressureLoss0D,
    PressureLossSys,
    PressureLossFamily,
    FalseSystem,
    Tank,
    Splitter12,
    Mixer21,
)
from .vectors import (
    Strait1dLine,
    Strait2dLine,
    Splitter1d,
    Merger1d,
    AllTypesSystem,
    BooleanSystem,
)



import pytest

from cosapp.ports import Port
from cosapp.systems import System

from ..markdown import system_to_md

# <codecell> Local test classes

class DummyPort(Port):
    def setup(self):
        self.add_variable("a", 1)
        self.add_variable("b", 2)

class AnotherPort(Port):
    def setup(self):
        self.add_variable("aaaa", 1)

class System2(System):
    def setup(self):
        self.add_inward({"data1": 9.0, "data2": 11.0, "data3": 13.0})
        self.add_outward({"local1": 7.0, "a": 14.0, "b": 21.0})
        self.add_output(AnotherPort, "other")

class System3(System):
    def setup(self):
        self.add_input(DummyPort, "entry")
        self.add_output(DummyPort, "exit")

class System4(System):
    def setup(self):
        self.add_inward("X", -1.0)
        self.add_equation("X == 1.")

class DynamicSystem2(System2):
    def setup(self):
        super().setup()
        self.add_transient("A", der="a")
        self.add_rate("db_dt", source="b")

class DynamicSystem3(System3):
    def setup(self):
        super().setup()
        self.add_transient("A", der="entry.a")
        self.add_transient("B", der="entry.b")
        self.add_rate("db_out", source="exit.b")

# <codecell>

# TODO('be sure the test is complete')
def test_markdown():
    s = System2("s")
    markdown = system_to_md(s)
    assert "### Child components" not in markdown
    assert "#### {}".format(System.INWARDS.capitalize()) in markdown
    assert "#### {}".format(System.OUTWARDS.capitalize()) in markdown
    assert "### Ports" in markdown
    assert "#### Inputs" not in markdown
    assert "#### Outputs" in markdown
    assert "#### Residues" not in markdown

    s = System3("s")
    markdown = system_to_md(s)
    assert "### Child components" not in markdown
    assert "#### {}".format(System.INWARDS.capitalize()) not in markdown
    assert "#### {}".format(System.OUTWARDS.capitalize()) not in markdown
    assert "### Ports" in markdown
    assert "#### Inputs" in markdown
    assert "#### Outputs" in markdown
    assert "#### Residues" not in markdown

    s = System4("s")
    markdown = system_to_md(s)
    assert "### Child components" not in markdown
    assert "#### {}".format(System.INWARDS.capitalize()) in markdown
    assert "#### {}".format(System.OUTWARDS.capitalize()) not in markdown
    assert "### Ports" not in markdown
    assert "#### Inputs" not in markdown
    assert "#### Outputs" not in markdown
    assert "### Residues" in markdown

    u = System("child")
    s.add_child(u)
    markdown = system_to_md(s)
    assert "### Child components" in markdown

    class NewS(System):
        tags = ["dev"]

    s = NewS("s")
    markdown = system_to_md(s)
    assert "**Tags**" in markdown



import json

import pytest

from ..distribution import Distribution


class DummyDistribution(Distribution):
    def _set_distribution(self):
        pass

    def draw(self, q):
        return q


def test_Distribution___init__():
    distrib = DummyDistribution(3.0, 7.0)
    assert isinstance(distrib, Distribution)

    assert distrib.worst == 3.0
    assert distrib.best == 7.0
    assert distrib.pworst == 0.15
    assert distrib.pbest == 0.15

    distrib = DummyDistribution(12.0, 5.0, 0.2, 0.8)
    assert isinstance(distrib, Distribution)

    assert distrib.worst == 12.0
    assert distrib.best == 5.0
    assert distrib.pworst == 0.2
    assert distrib.pbest == 0.8


def test_Distribution___json__():
    distrib = DummyDistribution(3.0, 7.0, 0.1, 0.9)
    try:
        json.dumps(distrib.__json__())
    except TypeError:
        pytest.fail("Distribution.__json__ raised TypeError")

    assert distrib.__json__() == {
        "worst": 3.0,
        "pworst": 0.1,
        "best": 7.0,
        "pbest": 0.9,
    }


def test_Distribution_worst():
    distrib = DummyDistribution(3.0, 7.0)

    distrib.worst = 4
    assert distrib.worst == 4


def test_Distribution_best():
    distrib = DummyDistribution(3.0, 7.0)

    distrib.best = 6
    assert distrib.best == 6


def test_Distribution_pworst():
    distrib = DummyDistribution(3.0, 7.0)

    distrib.pworst = 0.2
    assert distrib.pworst == 0.2

    with pytest.raises(ValueError):
        distrib.pworst = -1

    with pytest.raises(ValueError):
        distrib.pworst = 1.5


def test_Distribution_pbest():
    distrib = DummyDistribution(3.0, 7.0)

    distrib.pbest = 0.2
    assert distrib.pbest == 0.2

    with pytest.raises(ValueError):
        distrib.pbest = -1

    with pytest.raises(ValueError):
        distrib.pbest = 1.5



"""This module define the basic class encapsulating a variable attributes."""
import array
import numpy
import logging
from collections.abc import MutableSequence
from numbers import Number
from typing import Any, Dict, Iterable, Optional, Tuple, Union, NoReturn


from cosapp.core.numerics.distributions.distribution import Distribution
from cosapp.ports import units
from cosapp.ports.enum import Scope, Validity, RangeType
from cosapp.utils.helpers import check_arg, is_numerical, get_typename
from cosapp.utils.naming import NameChecker

logger = logging.getLogger(__name__)

RangeValue = Optional[Tuple[Any, Any]]
ArrayIndices = Optional[Union[int, Iterable[int], Iterable[Iterable[int]], slice]]
Types = Optional[Union[Any, Tuple[Any, ...]]]


class Variable:
    """Variable details container.

    The `valid_range` defines the range of value for which a model is known to behave correctly.
    The `limits` are at least as large as the validity range.

    Parameters
    ----------
    name : str
        Variable name.
    port : ExtensiblePort
        Port to which belong the variable
    value : Any
        Variable value
    unit : str, optional
        Variable unit; default empty string (i.e. dimensionless)
    dtype : type or iterable of type, optional
        Variable type; default None (i.e. type of initial value)
    valid_range : Tuple[Any, Any] or Tuple[Tuple], optional
        Validity range of the variable; default None (i.e. all values are valid).
        Tuple[Any, Any] in case of scalar value, tuple of tuples in case of vector value.
    invalid_comment : str, optional
        Comment to show in case the value is not valid; default ''
    limits : Tuple[Any, Any] or Tuple[Tuple], optional
        Limits over which the use of the model is wrong; default valid_range.
        Tuple[Any, Any] in case of scalar value, tuple of tuples in case of vector value.
    out_of_limits_comment : str, optional
        Comment to show in case the value is not valid; default ''
    desc : str, optional
        Variable description; default ''
    distribution : Distribution, optional
        Variable random distribution; default None (no distribution)
    scope : Scope {PRIVATE, PROTECTED, PUBLIC}, optional
        Variable visibility; default PRIVATE
    """

    # Value cannot be integrated in this object
    #   otherwise value getter execution time *200
    #   another advantage is that the metadata can be change/shared without providing
    #     a way to change the value no passing through the Port (which is bad for the
    #     clean dirty logic).
    #   Disadvantage : the logic to store the details and the value is a bit messy in
    #     ExtensiblePort.add_variable

    __slots__ = [
        "__weakref__",
        "_desc",
        "_distribution",
        "_invalid_comment",
        "_limits",
        "_name",
        "_port",
        "_out_of_limits_comment",
        "_scope",
        "_dtype",
        "_unit",
        "_valid_range",
    ]

    __name_check = NameChecker(excluded=["inwards", "outwards"])

    @staticmethod
    def _get_limits_from_type(variable: Any) -> RangeValue:
        """Get default limits for a variable depending of its type.

        Parameters
        ----------
        variable : Any
            Variable of interest

        Returns
        -------
        Tuple[float, float] or None
            Default (lower, upper) limits
        """
        if is_numerical(variable):
            return -numpy.inf, numpy.inf
        else:
            return None

    @property
    def full_name(self) -> str:
        return f"{self._port.contextual_name}.{self.name}"

    def check_range_type(self, value: Iterable) -> RangeType:
        """Get type of valid_range of limits of variable.
        This function checks if `value` is a size 2 tuple of scalar
        or a tuple of (lower, upper) tuples.
        
        Parameters
        ----------
        value : Any
            value need to be checked

        Returns
        -------
        int (from enum RangeType)
            Type of `value`
        """
        tuple_check = True
        value_check = True

        if isinstance(value, Number):
            raise TypeError(
                "Validity or limit range must be a tuple with format comparable to value"
            )
        if value is not None:
            for bound in value :
                if isinstance(bound, (tuple,list)):
                    value_check = False
                elif isinstance(bound, Number) or bound is None: 
                    tuple_check = False
                else:
                    value_check = False
                    tuple_check = False        
        else :
            return RangeType.NONE
        if value_check and not tuple_check:
            if len(value) != 2:
                raise TypeError(
                    "Valid range or limits must be a size 2 tuple with type comparable to value"
                )
            return RangeType.VALUE
        elif tuple_check and not value_check:
            return RangeType.TUPLE
        else:
            raise ValueError(
                f"Mixed values in valid_range {value} of {self.full_name!r}."
                " Valid object can contain only numerical values or only tuples"
            ) 

    def _check_range(self,
        limits: RangeValue,
        valid_range: RangeValue,
        value: Any,
    ) -> Tuple[RangeValue, RangeValue]:
        """Correct coherence of limits and validation range depending on value type.

        Parameters
        ----------
        limits : Tuple[Any, Any] or Tuple[Tuple] or None
            (lower, upper) limits
        valid_range : Tuple[Any, Any] Tuple[Tuple] or None
            (lower, upper) validation range
        value : Any

        Returns
        -------
        Tuple[Tuple[Any, Any] or Tuple[Tuple] or None, Tuple[Any, Any] or Tuple[Tuple] or None]
            Tuple of corrected (limits, validation range)
        """
        default = Variable._get_limits_from_type(value)
        range_type = self.check_range_type(valid_range)
        limits_type = self.check_range_type(limits)

        if default is not None:

            def get_bounds(lower, upper) -> Tuple[float, float]:
                if lower is None:
                    lower = default[0]
                if upper is None:
                    upper = default[1]
                return (lower, upper) if lower < upper else (upper, lower)

            def raise_inconsistency() -> NoReturn:
                name = self.full_name
                raise ValueError(
                    f"valid_range {valid_range} and limits {limits} of variable {name!r} have different formats"
                )

            if limits is None:
                limits = (None, None)

            if valid_range is None:
                valid_range = limits

            if range_type == RangeType.VALUE:

                min_range, max_range = valid_range = get_bounds(*valid_range)

                if limits_type == RangeType.VALUE:
                    min_limit, max_limit = get_bounds(*limits)
                    limits = (numpy.minimum(min_range, min_limit), numpy.maximum(max_range, max_limit))
                elif limits_type == RangeType.NONE:
                    limits = tuple(default)
                else:
                    raise_inconsistency()

            elif range_type == RangeType.TUPLE:

                valid_range = tuple(get_bounds(*pair) for pair in valid_range)

                if limits_type == RangeType.TUPLE:
                    limits = tuple(get_bounds(*pair) for pair in limits)
                elif limits_type == RangeType.NONE:
                    limits = tuple(default)
                else:
                    raise_inconsistency()
                    
            elif range_type == RangeType.NONE:

                if limits_type == RangeType.VALUE:
                    limits = valid_range = get_bounds(*limits)
                elif limits_type == RangeType.TUPLE:
                    limits = valid_range = tuple(get_bounds(*pair) for pair in limits)
                elif limits_type == RangeType.NONE:
                    limits = valid_range = tuple(default)

        else:
            limits = valid_range = None

        return limits, valid_range

    def __init__(
        self,
        name: str,
        port: "cosapp.ports.port.ExtensiblePort",
        value: Any,
        unit: str = "",
        dtype: Types = None,
        valid_range: RangeValue = None,
        invalid_comment: str = "",
        limits: RangeValue = None,
        out_of_limits_comment: str = "",
        desc: str = "",
        distribution: Optional[Distribution] = None,
        scope: Scope = Scope.PRIVATE,
    ):
        self._name = self.name_check(name)

        from cosapp.ports.port import ExtensiblePort
        check_arg(port, 'port', ExtensiblePort)
        self._port = port

        check_arg(unit, 'unit', str)
        if dtype is not None:
            if not isinstance(dtype, type):
                failed = False
                try:
                    for dtype_ in dtype:
                        failed |= not isinstance(dtype_, type)
                except TypeError:
                    failed = True
                if failed:
                    raise TypeError(f"Types must be defined by a type; got {dtype}.")

        check_arg(desc, 'desc', str)
        check_arg(scope, 'scope', Scope)

        if unit and not is_numerical(value):
            unit = ""
            logger.warning(
                f"A physical unit is defined for non-numerical variable {name!r}; it will be ignored."
            )

        elif not units.is_valid_units(unit) and unit:
            raise units.UnitError(f"Unknown unit {unit}.")

        if dtype is None:
            if value is None:
                dtype = None  # can't figure out type if both value and dtype are None
            else:
                dtype = type(value)
                if is_numerical(value):
                    # Force generic number type only if user has not specified any type
                    if issubclass(dtype, Number):
                        dtype = (Number, numpy.ndarray)
                    elif isinstance(value, (MutableSequence, array.ArrayType)):
                        # We have a collection => transform Mutable to ndarray
                        dtype = (MutableSequence, array.ArrayType, numpy.ndarray)
                        value = numpy.asarray(value)
        elif value is not None:
            # Test value has the right type
            if not isinstance(value, dtype):
                typename = get_typename(dtype)
                varname = f"{port.contextual_name}.{name}"
                raise TypeError(
                    "Cannot set {} of type {} with a {}.".format(
                        varname, typename, type(value).__qualname__
                    )
                )

        # TODO: better handle of this possible misunderstanding for users at numpy array instantiation
        if isinstance(value, numpy.ndarray):
            if issubclass(value.dtype.type, numpy.integer):
                logger.warning(
                    f"Variable {name!r} instantiates a numpy array with integer dtype."
                    " This may lead to unpredictible consequences."
                )

        # Check validation ranges are compatible and meaningful for this type of data
        limits, valid_range = self._check_range(limits, valid_range, value)

        if valid_range is None and len(invalid_comment) > 0:
            logger.warning(
                f"Invalid comment specified for variable {name!r} without validity range."
            )

        if limits is None and len(out_of_limits_comment) > 0:
            logger.warning(
                f"Out-of-limits comment specified for variable {name!r} without limits."
            )

        self._unit = unit  # type: str
        self._dtype = dtype  # type: Types
        self._valid_range = valid_range  # type: RangeValue
        self._invalid_comment = ""  # type: str
        self.invalid_comment = invalid_comment
        self._limits = limits  # type: RangeValue
        self._out_of_limits_comment = ""  # type: str
        self.out_of_limits_comment = out_of_limits_comment
        self._desc = desc  # type: str
        self._distribution = None  # type: Optional[Distribution]
        self.distribution = distribution
        self._scope = scope  # type: Scope

    @classmethod
    def name_check(cls, name: str):
        return cls.__name_check(name)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        msg = {"name": self.name, "unit": f" {self.unit}" if self.unit else ""}
        value = self.value
        try:
            msg["value"] = f"{value:.5g}"
        except:
            msg["value"] = value

        min_valid, max_valid = (
            self.valid_range if self.valid_range is not None else (None, None)
        )
        min_limit, max_limit = self.limits if self.limits is not None else (None, None)
        if min_limit is None or numpy.all(numpy.isinf(min_limit)):
            msg["min_limit"] = ""
        else:
            msg["min_limit"] = f" &#10647; {min_limit:.5g} &#10205; "
        if max_limit is None or numpy.all(numpy.isinf(max_limit)):
            msg["max_limit"] = ""
        else:
            msg["max_limit"] = f" &#10206; {max_limit:.5g} &#10648; "
        if min_valid is None or min_limit == min_valid or numpy.all(numpy.isinf(min_valid)):
            msg["min_valid"] = ""
        else:
            msg["min_valid"] = f"{min_valid:.5g} &#10205; "
        if max_valid is None or max_limit == max_valid or numpy.all(numpy.isinf(max_valid)):
            msg["max_valid"] = ""
        else:
            msg["max_valid"] = f" &#10206; {max_valid:.5g}"
        if self.description:
            msg["description"] = f" # {self.description}"
        else:
            msg["description"] = ""

        scope_format = {
            Scope.PRIVATE: " &#128274;&#128274; ",
            Scope.PROTECTED: " &#128274; ",
            Scope.PUBLIC: "",
        }
        msg["scope"] = scope_format[self.scope]

        if len(msg["min_limit"] + msg["min_valid"]) == len(msg["max_limit"] + msg["max_valid"]) == 0:
            msg["range"] = ""
            msg["separator"] = ""
        else:
            msg["range"] = " value "
            msg["separator"] = "; "

        return (
            "{name}{scope}: {value!s}{unit}"
            "{separator}{min_limit}{min_valid}{range}{max_valid}{max_limit}"
            "{description}".format(**msg)
        )

    def __json__(self) -> Dict[str, Any]:
        """JSONable dictionary representing a variable.
        
        Returns
        -------
        Dict[str, Any]
            The dictionary
        """
        return {
            "value": self.value,
            "valid_range": self.valid_range,
            "invalid_comment": self.invalid_comment,
            "limits": self.limits,
            "out_of_limits_comment": self.out_of_limits_comment,
            "distribution": self.distribution.__json__() if self.distribution else None,
        }

    @property
    def name(self) -> str:
        """str : Variable name"""
        return self._name

    @property
    def value(self) -> Any:
        return getattr(self._port, self._name)

    @property
    def unit(self) -> str:
        """str : Variable unit; empty string means dimensionless"""
        return self._unit

    @property
    def dtype(self) -> Types:
        """Type[Any] or Tuple of Type[Any] or None : Type of the variable; default None (i.e. type of default value is set)"""
        return self._dtype

    @property
    def valid_range(self) -> RangeValue:
        """Tuple[Any, Any] or None : alidity range of the variable and optional comment if unvalid"""
        
        return self._valid_range

    @valid_range.setter
    def valid_range(self, new_range: RangeValue):
        
        range_type = self.check_range_type(new_range)

        value = self.value
        default = self._get_limits_from_type(value)

        if default is not None:
            if range_type == RangeType.VALUE :
                if len(new_range) != 2:
                    raise TypeError(
                        "Validity range must be a size 2 tuple with type comparable to value."
                    )
                limits = self.limits
                valid_range = new_range

            elif range_type == RangeType.TUPLE: 
                limits = self.limits
                valid_range = new_range
                current_range = self.valid_range
                 
            elif range_type == RangeType.NONE:
                limits = self.limits
                valid_range = new_range
                current_range = self.valid_range
                min_valid = (
                    limits[0] if limits[0] > current_range[0] else current_range[0]
                )
                max_valid = (
                    limits[1] if limits[1] < current_range[1] else current_range[1]
                )
                valid_range = (min_valid, max_valid)      
        

        else:
            valid_range = None
            limits = None

        limits, valid_range = self._check_range(self.limits, new_range, value)
        self._valid_range = valid_range
        self._limits = limits

    @property
    def invalid_comment(self) -> str:
        """str : Comment explaining the reasons of the validity range"""
        return self._invalid_comment

    @invalid_comment.setter
    def invalid_comment(self, new_comment: str):
        if not isinstance(new_comment, str):
            raise TypeError(f"invalid_comment must be a string; got {new_comment!s}")
        self._invalid_comment = new_comment

    @property
    def limits(self) -> RangeValue:
        """Tuple[Any, Any] or None : Variable limits and optional comment if unvalid"""
        return self._limits

    @limits.setter
    def limits(self, new_limits: RangeValue):

        limits_type = self.check_range_type(new_limits)

        value = self.value
        default = self._get_limits_from_type(value)

        if default is not None:
            if limits_type == RangeType.VALUE :
                if len(new_limits) != 2:
                    raise TypeError(
                        "Limits must be a size 2 tuple with type comparable to value."
                    )

                limits = new_limits
                if new_limits[0] is None:
                    limits = (default[0], limits[1])
                if new_limits[1] is None:
                    limits = (limits[0], default[1])

                current_range = self.valid_range
                min_valid = max(limits[0], current_range[0])
                max_valid = min(limits[1], current_range[1])
                valid_range = (min_valid, max_valid)
            
            elif limits_type == RangeType.TUPLE:
                limits_list = list(new_limits)
                for index in range(len(limits_list)):

                    if limits_list[index][0] is None:
                       limits_list[index] =  (default[0], limits_list[index][1])
                    if limits_list[index][1] is None:
                       limits_list[index][1] =  (limits_list[index][0],default[1])
                limits = tuple(limits_list)

                current_range = self.valid_range
                valid_range_list = list(current_range)
                for index, valid_range in enumerate(valid_range_list):
                    limit_value = limits[index]
                    range_value = current_range[index]
                    valid_range_list[index] = (max(limit_value[0], range_value[0]), valid_range[1]) 
                    valid_range_list[index] = (valid_range[0], min(limit_value[1], range_value[1]))
                valid_range = tuple(valid_range_list)

            elif limits_type == RangeType.NONE:
                limits = self.limits
                current_range = self.valid_range
                min_valid = max(limits[0], current_range[0])
                max_valid = min(limits[1], current_range[1])
                valid_range = (min_valid, max_valid)


        else:
            limits = valid_range = None

        limits, valid_range = self._check_range(limits, valid_range, value)
        self._valid_range = valid_range
        self._limits = limits

    @property
    def out_of_limits_comment(self) -> str:
        """str : Comment explaining the reasons of the limits"""
        return self._out_of_limits_comment

    @out_of_limits_comment.setter
    def out_of_limits_comment(self, new_comment: str):
        if not isinstance(new_comment, str):
            raise TypeError(f"out_of_limits_comment must be a string; got {new_comment!s}")
        self._out_of_limits_comment = new_comment

    @property
    def description(self) -> str:
        """str : Variable description"""
        return self._desc

    @property
    def scope(self) -> Scope:
        """Scope : Scope of variable visibility"""
        return self._scope

    @property
    def distribution(self) -> Optional[Distribution]:
        """Optional[Distribution] : Random distribution of the variable."""
        return self._distribution

    @distribution.setter
    def distribution(self, new_distribution: Optional[Distribution]):
        ok = new_distribution is None or isinstance(new_distribution, Distribution)
        if not ok:
            typename = type(new_distribution).__qualname__
            raise TypeError(f"Random distribution should be of type 'Distribution'; got {typename}.")
        self._distribution = new_distribution

    def is_valid(self) -> Validity:
        """Get the variable value validity.

        Returns
        -------
        Validity
            Variable value validity
        """
        status = Validity.OK
        value = self.value

        if not isinstance(value, (Number, numpy.ndarray)):
            return status

        if self.valid_range is not None:

            range_type = self.check_range_type(self.valid_range)
            if isinstance(value, numpy.ndarray):  
            
                if range_type == RangeType.VALUE:
                    min_range, max_range = self.valid_range 
                    if numpy.any(value > max_range) or numpy.any(value < min_range):
                        status = Validity.WARNING
                    
                    if self.limits is not None:
                        min_limit, max_limit = self.limits
                        if numpy.any(value > max_limit) or numpy.any(value < min_limit):
                            status = Validity.ERROR

                elif range_type == RangeType.TUPLE:

                    def check_values(bound_list, key) -> None:
                        nonlocal status
                        for v, (lower, upper) in zip(value, bound_list):
                            if not (lower <= v <= upper):
                                status = Validity[key]
                                break

                    check_values(self.valid_range, "WARNING")
                    
                    if self.limits is not None:
                        check_values(self.limits, "ERROR")

                else:
                    varname = f"{self._port.contextual_name}.{self.name}"
                    raise ValueError(
                        f"Mixed values in valid_range {self.valid_range} of {varname!r}."
                        " Valid object can contain only numerical values or only tuples")                     
                
            else:    
                if range_type == RangeType.VALUE:
                    min_range, max_range = self.valid_range          
                    if not min_range <= value <= max_range:
                        status = Validity.WARNING

                        if self.limits is not None:
                            min_limit, max_limit = self.limits
                            if not min_limit <= value <= max_limit:
                                status = Validity.ERROR

                elif range_type == RangeType.TUPLE:
                    varname = f"{self._port.contextual_name}.{self.name}"
                    raise ValueError(
                        f"valid_range {self.valid_range} or limits {self.limits} of variable {varname!r}"
                        f" are incompatible with its value {value}")

        return status

    def get_validity_comment(self, status: Validity) -> str:
        """Get the ground arguments used to established the variable validity range.

        The status `Validity.OK` has no ground arguments.

        Parameters
        ----------
        status : Validity.{OK, WARNING, ERROR}
            Validity status for which the reasons are looked for.

        Returns
        -------
        str
            Validity comment
        """

        def range2str(range: RangeValue) -> str:

            range_type = self.check_range_type(range)

            if range_type == RangeType.VALUE:
                min_valid, max_valid = range if range is not None else (None, None)

            elif range_type == RangeType.TUPLE:
                return f"{range}"[1:-1].join("[]")

            elif range_type == RangeType.NONE:
                min_valid, max_valid = (None, None)


            def get_range_repr(valid, fmt):
                if not isinstance(valid, (list, tuple, numpy.ndarray)):
                    fmt.replace("{}", "{:.5g}")
                return fmt.format(valid)

            if min_valid is None:
                range_repr = "] ,"
            else:
                range_repr = get_range_repr(min_valid, "[{}, ")

            if max_valid is None:
                range_repr += " ["
            else:
                range_repr += get_range_repr(max_valid, "{}]")
            return range_repr

        if status == Validity.ERROR:
            return f"{range2str(self.limits)} - {self.out_of_limits_comment}"
        elif status == Validity.WARNING:
            return f"{range2str(self.valid_range)} - {self.invalid_comment}"
        else:  # Variable is ok
            return ""

    def to_dict(self) -> Dict:
        """Convert this variable into a dictionary.
   
        Returns
        -------
        dict
            The dictionary representing this variable.
        """
        ret = {
            "value" : self.value,
            "unit": self.unit or None,
            "invalid_comment": self.invalid_comment or None,
            "out_of_limits_comment": self.out_of_limits_comment or None,
            "desc" : self.description or None,
            "distribution": self.distribution.__json__() if self.distribution else None,
        } 
        
        for key in ["valid_range", "limits"]:
            tmp_val = list(getattr(self, key))
            for idx, val in enumerate(tmp_val):
                if numpy.isinf(val):
                    tmp_val[idx] = str(val) 
            if  tmp_val == ["-inf", "inf"]:
                tmp_val = None
            ret[key] = tmp_val 
        
        return { key: value for (key, value) in ret.items() if value is not None }



"""
Integrates the Multi-Fidelity Co-Kriging method described in [LeGratiet2013].

(Author: Remi Vauclin vauclin.remi@gmail.com)

This code was implemented using the package scikit-learn as basis.
(Author: Vincent Dubourg, vincent.dubourg@gmail.com)

OpenMDAO adaptation. Regression and correlation functions were directly copied
from scikit-learn package here to avoid scikit-learn dependency.
(Author: Remi Lafage, remi.lafage@onera.fr)

ISAE/DMSM - ONERA/DCPS
"""

import numpy as np
from numpy import atleast_2d as array2d

from scipy import linalg
from scipy.optimize import minimize
from scipy.spatial.distance import squareform

from .surrogate_model import MultiFiSurrogateModel

import logging
_logger = logging.getLogger()

MACHINE_EPSILON = np.finfo(np.double).eps  # machine precision
NUGGET = 10. * MACHINE_EPSILON  # nugget for robustness

INITIAL_RANGE_DEFAULT = 0.3  # initial range for optimizer
TOLERANCE_DEFAULT = 1e-6    # stopping criterion for MLE optimization

THETA0_DEFAULT = 0.5
THETAL_DEFAULT = 1e-5
THETAU_DEFAULT = 50

if hasattr(linalg, 'solve_triangular'):
    # only in scipy since 0.9
    solve_triangular = linalg.solve_triangular
else:
    # slower, but works
    def solve_triangular(x, y, lower=True):
        """Solve triangular."""
        return linalg.solve(x, y)


def constant_regression(x):
    """
    Zero order polynomial (constant, p = 1) regression model.

    x --> f(x) = 1

    Parameters
    ----------
    x : array_like
        Input data.

    Returns
    -------
    array_like
        Constant regression output.
    """
    x = np.asarray(x, dtype=float)
    n_eval = x.shape[0]
    f = np.ones([n_eval, 1])
    return f


def linear_regression(x):
    """
    First order polynomial (linear, p = n+1) regression model.

    x --> f(x) = [ 1, x_1, ..., x_n ].T

    Parameters
    ----------
    x : array_like
        Input data.

    Returns
    -------
    array_like
        Linear regression output.
    """
    x = np.asarray(x, dtype=float)
    n_eval = x.shape[0]
    f = np.hstack([np.ones([n_eval, 1]), x])
    return f


def squared_exponential_correlation(theta, d):
    """
    Squared exponential correlation model (Radial Basis Function).

    (Infinitely differentiable stochastic process, very smooth)::

                                            n
        theta, dx --> r(theta, dx) = exp(  sum  - theta_i * (dx_i)^2 )
                                          i = 1

    Parameters
    ----------
    theta : array_like
        An array with shape 1 (isotropic) or n (anisotropic) giving the
        autocorrelation parameter(s).
    d : array_like
        An array with shape (n_eval, n_features) giving the componentwise
        distances between locations x and x' at which the correlation model
        should be evaluated.

    Returns
    -------
    r : array_like
        An array with shape (n_eval, ) containing the values of the
        autocorrelation model.
    """
    theta = np.asarray(theta, dtype=float)
    d = np.asarray(d, dtype=float)

    if d.ndim > 1:
        n_features = d.shape[1]
    else:
        n_features = 1

    if theta.size == 1:
        return np.exp(-theta[0] * np.sum(d ** 2, axis=1))
    elif theta.size != n_features:
        raise ValueError("Length of theta must be 1 or %s" % n_features)
    else:
        return np.exp(-np.sum(theta.reshape(1, n_features) * d ** 2, axis=1))


def l1_cross_distances(X, Y=None):
    """
    Compute the nonzero componentwise L1 cross-distances between the vectors in X and Y.

    Parameters
    ----------
    X : array_like
        An array with shape (n_samples_X, n_features)
    Y : array_like
        An array with shape (n_samples_Y, n_features)

    Returns
    -------
    array with shape (n_samples * (n_samples - 1) / 2, n_features)
        The array of componentwise L1 cross-distances.

    """
    if Y is None:
        X = array2d(X)
        n_samples, n_features = X.shape
        n_nonzero_cross_dist = n_samples * (n_samples - 1) // 2
        D = np.zeros((n_nonzero_cross_dist, n_features))
        ll_1 = 0
        for k in range(n_samples - 1):
            ll_0 = ll_1
            ll_1 = ll_0 + n_samples - k - 1
            D[ll_0:ll_1] = np.abs(X[k] - X[(k + 1):])
    else:
        X = array2d(X)
        Y = array2d(Y)
        n_samples_X, n_features_X = X.shape
        n_samples_Y, n_features_Y = Y.shape
        if n_features_X != n_features_Y:
            raise ValueError("X and Y must have the same dimensions.")
        n_features = n_features_X

        n_nonzero_cross_dist = n_samples_X * n_samples_Y
        D = np.zeros((n_nonzero_cross_dist, n_features))
        ll_1 = 0
        for k in range(n_samples_X):
            ll_0 = ll_1
            ll_1 = ll_0 + n_samples_Y  # - k - 1
            D[ll_0:ll_1] = np.abs(X[k] - Y)

    return D


class MultiFiCoKriging(object):
    """
    Integrate the Multi-Fidelity Co-Kriging method described in [LeGratiet2013].

    Attributes
    ----------
    corr : Object
        Correlation function to use, default is squared_exponential_correlation.
    regr : string or callable
        A regression function returning an array of outputs of the linear
        regression functional basis for Universal Kriging purpose.
        regr is assumed to be the same for all levels of code.
        Default assumes a simple constant regression trend.
        Available built-in regression models are:
        'constant', 'linear'
    rho_regr : string or callable or None
        A regression function returning an array of outputs of the linear
        regression functional basis. Defines the regression function for the
        autoregressive parameter rho.
        rho_regr is assumed to be the same for all levels of code.
        Default assumes a simple constant regression trend.
        Available built-in regression models are:
        'constant', 'linear'
    theta : double, array_like or list or None
        Value of correlation parameters if they are known; no optimization is run.
        Default is None, so that optimization is run.
        if double: value is replicated for all features and all levels.
        if array_like: an array with shape (n_features, ) for
        isotropic calculation. It is replicated for all levels.
        if list: a list of nlevel arrays specifying value for each level
    theta0 : double, array_like or list or None
        Starting point for the maximum likelihood estimation of the
        best set of parameters.
        Default is None and meaning use of the default 0.5*np.ones(n_features)
        if double: value is replicated for all features and all levels.
        if array_like: an array with shape (n_features, ) for
        isotropic calculation. It is replicated for all levels.
        if list: a list of nlevel arrays specifying value for each level
    thetaL : double, array_like or list or None
        Lower bound on the autocorrelation parameters for maximum
        likelihood estimation.
        Default is None meaning use of the default 1e-5*np.ones(n_features).
        if double: value is replicated for all features and all levels.
        if array_like: An array with shape matching theta0's. It is replicated
        for all levels of code.
        if list: a list of nlevel arrays specifying value for each level
    thetaU : double, array_like or list or None
        Upper bound on the autocorrelation parameters for maximum
        likelihood estimation.
        Default is None meaning use of default value 50*np.ones(n_features).
        if double: value is replicated for all features and all levels.
        if array_like: An array with shape matching theta0's. It is replicated
        for all levels of code.
        if list: a list of nlevel arrays specifying value for each level
    _nfev : int
        Number of function evaluations.

    Examples
    --------
    >>> from cosapp.surrogate_models.multifi_cokriging import MultiFiCoKriging
    >>> import numpy as np
    >>> # Xe: DOE for expensive code (nested in Xc)
    >>> # Xc: DOE for cheap code
    >>> # ye: expensive response
    >>> # yc: cheap response
    >>> Xe = np.array([[0],[0.4],[1]])
    >>> Xc = np.vstack((np.array([[0.1],[0.2],[0.3],[0.5],[0.6],[0.7],[0.8],[0.9]]),Xe))
    >>> ye = ((Xe*6-2)**2)*np.sin((Xe*6-2)*2)
    >>> yc = 0.5*((Xc*6-2)**2)*np.sin((Xc*6-2)*2)+(Xc-0.5)*10. - 5
    >>> model = MultiFiCoKriging(theta0=1, thetaL=1e-5, thetaU=50.)
    >>> model.fit([Xc, Xe], [yc, ye])
    >>> # Prediction on x=0.05
    >>> np.abs(float(model.predict([0.05])[0])- ((0.05*6-2)**2)*np.sin((0.05*6-2)*2)) < 0.05
    True


    Notes
    -----
    Implementation is based on the Package Scikit-Learn
    (Author: Vincent Dubourg, vincent.dubourg@gmail.com) which translates
    the DACE Matlab toolbox, see [NLNS2002]_.


    References
    ----------
    .. [NLNS2002] H. B. Nielsen, S. N. Lophaven, and J. Sondergaard.
       `DACE - A MATLAB Kriging Toolbox.` (2002)
       http://www2.imm.dtu.dk/~hbn/dace/dace.pdf

    .. [WBSWM1992] W. J. Welch, R. J. Buck, J. Sacks, H. P. Wynn, T. J. Mitchell,
       and M. D. Morris (1992). "Screening, predicting, and computer experiments."
       `Technometrics,` 34(1) 15--25.
       http://www.jstor.org/pss/1269548

    .. [LeGratiet2013] L. Le Gratiet (2013). "Multi-fidelity Gaussian process
       regression for computer experiments."
       PhD thesis, Universite Paris-Diderot-Paris VII.

    .. [TBKH2011] Toal, D. J., Bressloff, N. W., Keane, A. J., & Holden, C. M. E. (2011).
       "The development of a hybridized particle swarm for kriging hyperparameter
       tuning." `Engineering optimization`, 43(6), 675-699.
    """

    _regression_types = {
        'constant': constant_regression,
        'linear': linear_regression
    }

    def __init__(self, regr='constant', rho_regr='constant',
                 theta=None, theta0=None, thetaL=None, thetaU=None):
        """
        Initialize all attributes.

        Parameters
        ----------
        regr : string or callable, optional
            A regression function returning an array of outputs of the linear
            regression functional basis for Universal Kriging purpose.
            regr is assumed to be the same for all levels of code.
            Default assumes a simple constant regression trend.
            Available built-in regression models are:
            'constant', 'linear'
        rho_regr : string or callable, optional
            A regression function returning an array of outputs of the linear
            regression functional basis. Defines the regression function for the
            autoregressive parameter rho.
            rho_regr is assumed to be the same for all levels of code.
            Default assumes a simple constant regression trend.
            Available built-in regression models are:
            'constant', 'linear'
        theta : double, array_like or list, optional
            Value of correlation parameters if they are known; no optimization is run.
            Default is None, so that optimization is run.
            if double: value is replicated for all features and all levels.
            if array_like: an array with shape (n_features, ) for
            isotropic calculation. It is replicated for all levels.
            if list: a list of nlevel arrays specifying value for each level
        theta0 : double, array_like or list, optional
            Starting point for the maximum likelihood estimation of the
            best set of parameters.
            Default is None and meaning use of the default 0.5*np.ones(n_features)
            if double: value is replicated for all features and all levels.
            if array_like: an array with shape (n_features, ) for
            isotropic calculation. It is replicated for all levels.
            if list: a list of nlevel arrays specifying value for each level
        thetaL : double, array_like or list, optional
            Lower bound on the autocorrelation parameters for maximum
            likelihood estimation.
            Default is None meaning use of the default 1e-5*np.ones(n_features).
            if double: value is replicated for all features and all levels.
            if array_like: An array with shape matching theta0's. It is replicated
            for all levels of code.
            if list: a list of nlevel arrays specifying value for each level
        thetaU : double, array_like or list, optional
            Upper bound on the autocorrelation parameters for maximum
            likelihood estimation.
            Default is None meaning use of default value 50*np.ones(n_features).
            if double: value is replicated for all features and all levels.
            if array_like: An array with shape matching theta0's. It is replicated
            for all levels of code.
            if list: a list of nlevel arrays specifying value for each level
        """
        self.corr = squared_exponential_correlation
        self.regr = regr
        self.rho_regr = rho_regr
        self.theta = theta
        self.theta0 = theta0
        self.thetaL = thetaL
        self.thetaU = thetaU

        self._nfev = 0

    def _build_R(self, lvl, theta):
        """
        Build the correlation matrix with given theta for the specified level.

        Parameters
        ----------
        lvl : Integer
            Level of fidelity
        theta : array_like
            An array containing the autocorrelation parameters at which the
            Gaussian Process model parameters should be determined.
            Default uses the built-in autocorrelation parameters
            (ie ``theta = self.theta``).

        Returns
        -------
        ndarray
            Correlatioin matrix.
        """
        D = self.D[lvl]
        n_samples = self.n_samples[lvl]

        R = np.eye(n_samples) * (1. + NUGGET)

        corr = squareform(self.corr(theta, D))
        R = R + corr

        return R

    def fit(self, X, y, initial_range=INITIAL_RANGE_DEFAULT, tol=TOLERANCE_DEFAULT):
        """
        Implement the Multi-Fidelity co-kriging model fitting method.

        Parameters
        ----------
        X : list of double array_like elements
            A list of arrays with the input at which observations were made, from lowest
            fidelity to highest fidelity. Designs must be nested
            with X[i] = np.vstack([..., X[i+1])
        y : list of double array_like elements
            A list of arrays with the observations of the scalar output to be predicted,
            from lowest fidelity to highest fidelity.
        initial_range : float
            Initial range for the optimizer.
        tol : float
            Optimizer terminates when the tolerance tol is reached.
        """
        # Run input checks
        # Transforms floats and arrays in lists to have a multifidelity
        # structure
        self._check_list_structure(X, y)
        # Checks if all parameters are structured as required
        self._check_params()

        X = self.X
        y = self.y
        nlevel = self.nlevel
        n_samples = self.n_samples

        # initialize lists
        self.beta = nlevel * [0]
        self.beta_rho = nlevel * [None]
        self.beta_regr = nlevel * [None]
        self.C = nlevel * [0]
        self.D = nlevel * [0]
        self.F = nlevel * [0]
        self.p = nlevel * [0]
        self.q = nlevel * [0]
        self.G = nlevel * [0]
        self.sigma2 = nlevel * [0]
        self._R_adj = nlevel * [None]

        y_best = y[nlevel - 1]
        for i in range(nlevel - 2, -1, -1):
            y_best = np.concatenate((y[i][:-n_samples[i + 1]], y_best))
        self.y_best = y_best

        self.y_mean = np.zeros(1)
        self.y_std = np.ones(1)
        self.X_mean = np.zeros(1)
        self.X_std = np.ones(1)

        for lvl in range(nlevel):

            # Calculate matrix of distances D between samples
            self.D[lvl] = l1_cross_distances(X[lvl])
            if (np.min(np.sum(self.D[lvl], axis=1)) == 0.):
                raise Exception("Multiple input features cannot have the same"
                                " value.")

            # Regression matrix and parameters
            self.F[lvl] = self.regr(X[lvl])
            self.p[lvl] = self.F[lvl].shape[1]

            # Concatenate the autoregressive part for levels > 0
            if lvl > 0:
                F_rho = self.rho_regr(X[lvl])
                self.q[lvl] = F_rho.shape[1]
                self.F[lvl] = np.hstack((F_rho * np.dot((self.y[lvl - 1])[-n_samples[lvl]:],
                                                        np.ones((1, self.q[lvl]))), self.F[lvl]))
            else:
                self.q[lvl] = 0

            n_samples_F_i = self.F[lvl].shape[0]

            if n_samples_F_i != n_samples[lvl]:
                raise Exception("Number of rows in F and X do not match. Most "
                                "likely something is going wrong with the "
                                "regression model.")

            if int(self.p[lvl] + self.q[lvl]) >= n_samples_F_i:
                raise Exception(("Ordinary least squares problem is undetermined "
                                 "n_samples=%d must be greater than the regression"
                                 " model size p+q=%d.")
                                % (n_samples[i], self.p[lvl] + self.q[lvl]))

        # Set attributes
        self.X = X
        self.y = y

        self.rlf_value = np.zeros(nlevel)

        for lvl in range(nlevel):
            # Determine Gaussian Process model parameters
            if self.theta[lvl] is None:
                # Maximum Likelihood Estimation of the parameters
                sol = self._max_rlf(
                    lvl=lvl, initial_range=initial_range, tol=tol)
                self.theta[lvl] = sol['theta']
                self.rlf_value[lvl] = sol['rlf_value']

                if np.isinf(self.rlf_value[lvl]):
                    raise Exception("Bad parameter region. "
                                    "Try increasing upper bound")
            else:
                self.rlf_value[lvl] = self.rlf(lvl=lvl)
                if np.isinf(self.rlf_value[lvl]):
                    raise Exception("Bad point. Try increasing theta0.")

        return

    def rlf(self, lvl, theta=None):
        """
        Determine BLUP parameters and evaluate negative reduced likelihood function for theta.

        Maximizing this function wrt the autocorrelation parameters theta is
        equivalent to maximizing the likelihood of the assumed joint Gaussian
        distribution of the observations y evaluated onto the design of
        experiments X.

        Parameters
        ----------
        lvl : Integer
            Level of fidelity
        theta : array_like, optional
            An array containing the autocorrelation parameters at which the
            Gaussian Process model parameters should be determined.
            Default uses the built-in autocorrelation parameters
            (ie ``theta = self.theta``).

        Returns
        -------
        double
            The value of the negative concentrated reduced likelihood function
            associated to the given autocorrelation parameters theta.
        """
        if theta is None:
            # Use built-in autocorrelation parameters
            theta = self.theta[lvl]

        # Initialize output
        rlf_value = 1e20

        # Retrieve data
        n_samples = self.n_samples[lvl]
        y = self.y[lvl]
        F = self.F[lvl]
        p = self.p[lvl]
        q = self.q[lvl]

        R = self._build_R(lvl, theta)

        try:
            C = linalg.cholesky(R, lower=True)
        except linalg.LinAlgError:
            _logger.warning(('Cholesky decomposition of R at level %i failed' % lvl) +
                            ' with theta=' + str(theta))
            return rlf_value

        # Get generalized least squares solution
        Ft = solve_triangular(C, F, lower=True)
        Yt = solve_triangular(C, y, lower=True)
        try:
            Q, G = linalg.qr(Ft, econ=True)
        except TypeError:   # qr() got an unexpected keyword argument 'econ'
            # DeprecationWarning: qr econ argument will be removed after scipy
            # 0.7. The economy transform will then be available through the
            # mode='economic' argument.
            Q, G = linalg.qr(Ft, mode='economic')
            pass

        # Universal Kriging
        beta = solve_triangular(G, np.dot(Q.T, Yt))

        err = Yt - np.dot(Ft, beta)
        err2 = np.dot(err.T, err)[0, 0]
        self._err = err
        sigma2 = err2 / (n_samples - p - q)
        detR = ((np.diag(C))**(2. / n_samples)).prod()

        rlf_value = (n_samples - p - q) * np.log10(sigma2) \
            + n_samples * np.log10(detR)

        self.beta_rho[lvl] = beta[:q]
        self.beta_regr[lvl] = beta[q:]
        self.beta[lvl] = beta
        self.sigma2[lvl] = sigma2
        self.C[lvl] = C
        self.G[lvl] = G

        return rlf_value

    def _max_rlf(self, lvl, initial_range, tol):
        """
        Estimate autocorrelation parameter theta as maximizer of the reduced likelihood function.

        (Minimization of the negative reduced likelihood function is used for convenience.)

        Parameters
        ----------
        lvl : integer
            Level of fidelity
        initial_range : float
            Initial range of the optimizer
        tol : float
            Optimizer terminates when the tolerance tol is reached.

        Returns
        -------
        array_like
            The optimal hyperparameters.
        double
            The optimal negative reduced likelihood function value.
        dict
            res['theta']: optimal theta
            res['rlf_value']: optimal value for likelihood
        """
        # Initialize input
        thetaL = self.thetaL[lvl]
        thetaU = self.thetaU[lvl]

        def rlf_transform(x):
            return self.rlf(theta=10.**x, lvl=lvl)

        # Use specified starting point as first guess
        theta0 = self.theta0[lvl]
        x0 = np.log10(theta0[0])

        constraints = []
        for i in range(theta0.size):
            constraints.append({'type': 'ineq', 'fun': lambda log10t, i=i:
                                log10t[i] - np.log10(thetaL[0][i])})
            constraints.append({'type': 'ineq', 'fun': lambda log10t, i=i:
                                np.log10(thetaU[0][i]) - log10t[i]})

        constraints = tuple(constraints)
        sol = minimize(rlf_transform, x0, method='COBYLA',
                       constraints=constraints,
                       options={'rhobeg': initial_range,
                                'tol': tol, 'disp': 0})

        log10_optimal_x = sol['x']
        optimal_rlf_value = sol['fun']
        self._nfev += sol['nfev']

        optimal_theta = 10. ** log10_optimal_x

        res = {}
        res['theta'] = optimal_theta
        res['rlf_value'] = optimal_rlf_value

        return res

    def predict(self, X, eval_MSE=True):
        """
        Perform the predictions of the kriging model on X.

        Parameters
        ----------
        X : array_like
            An array with shape (n_eval, n_features) giving the point(s) at
            which the prediction(s) should be made.
        eval_MSE : boolean, optional
            A boolean specifying whether the Mean Squared Error should be
            evaluated or not. Default assumes evalMSE is True.

        Returns
        -------
        array_like
            An array with shape (n_eval, ) with the Best Linear Unbiased
            Prediction at X. If all_levels is set to True, an array
            with shape (n_eval, nlevel) giving the BLUP for all levels.
        array_like, optional (if eval_MSE is True)
            An array with shape (n_eval, ) with the Mean Squared Error at X.
            If all_levels is set to True, an array with shape (n_eval, nlevel)
            giving the MSE for all levels.
        """
        X = array2d(X)
        nlevel = self.nlevel
        n_eval, n_features_X = X.shape

        # Calculate kriging mean and variance at level 0
        mu = np.zeros((n_eval, nlevel))

        f = self.regr(X)
        f0 = self.regr(X)
        dx = l1_cross_distances(X, Y=self.X[0])

        # Get regression function and correlation
        F = self.F[0]
        C = self.C[0]

        beta = self.beta[0]
        Ft = solve_triangular(C, F, lower=True)
        yt = solve_triangular(C, self.y[0], lower=True)
        r_ = self.corr(self.theta[0], dx).reshape(n_eval, self.n_samples[0])
        gamma = solve_triangular(C.T, yt - np.dot(Ft, beta), lower=False)

        # Scaled predictor
        mu[:, 0] = (np.dot(f, beta) + np.dot(r_, gamma)).ravel()

        if eval_MSE:
            self.sigma2_rho = nlevel * [None]
            MSE = np.zeros((n_eval, nlevel))
            r_t = solve_triangular(C, r_.T, lower=True)
            G = self.G[0]

            u_ = solve_triangular(G.T, f.T - np.dot(Ft.T, r_t), lower=True)
            MSE[:, 0] = self.sigma2[0] * \
                (1 - (r_t**2).sum(axis=0) + (u_**2).sum(axis=0))

        # Calculate recursively kriging mean and variance at level i
        for i in range(1, nlevel):
            C = self.C[i]
            F = self.F[i]
            g = self.rho_regr(X)
            dx = l1_cross_distances(X, Y=self.X[i])
            r_ = self.corr(self.theta[i], dx).reshape(
                n_eval, self.n_samples[i])
            f = np.vstack((g.T * mu[:, i - 1], f0.T))

            Ft = solve_triangular(C, F, lower=True)
            yt = solve_triangular(C, self.y[i], lower=True)
            r_t = solve_triangular(C, r_.T, lower=True)
            G = self.G[i]
            beta = self.beta[i]

            # scaled predictor
            mu[:, i] = (np.dot(f.T, beta)
                        + np.dot(r_t.T, yt - np.dot(Ft, beta))).ravel()

            if eval_MSE:
                Q_ = (np.dot((yt - np.dot(Ft, beta)).T,
                             yt - np.dot(Ft, beta)))[0, 0]
                u_ = solve_triangular(G.T, f - np.dot(Ft.T, r_t), lower=True)
                sigma2_rho = np.dot(g,
                                    self.sigma2[
                                        i] * linalg.inv(np.dot(G.T, G))[:self.q[i], :self.q[i]]
                                    + np.dot(beta[:self.q[i]], beta[:self.q[i]].T))
                sigma2_rho = (sigma2_rho * g).sum(axis=1)

                MSE[:, i] = sigma2_rho * MSE[:, i - 1] \
                    + Q_ / (2 * (self.n_samples[i] - self.p[i] - self.q[i])) \
                    * (1 - (r_t**2).sum(axis=0)) \
                    + self.sigma2[i] * (u_**2).sum(axis=0)

        # scaled predictor
        for i in range(nlevel):  # Predictor
            mu[:, i] = self.y_mean + self.y_std * mu[:, i]
            if eval_MSE:
                MSE[:, i] = self.y_std**2 * MSE[:, i]

        if eval_MSE:
            return mu[:, -1].reshape((n_eval, 1)), MSE[:, -1].reshape((n_eval, 1))
        else:
            return mu[:, -1].reshape((n_eval, 1))

    def _check_list_structure(self, X, y):
        """
        Transform floats and arrays in the training data lists to have a multifidelity structure.

        Parameters
        ----------
        X : list of double array_like elements
            A list of arrays with the input at which observations were made, from lowest
            fidelity to highest fidelity. Designs must be nested
            with X[i] = np.vstack([..., X[i+1])
        y : list of double array_like elements
            A list of arrays with the observations of the scalar output to be predicted,
            from lowest fidelity to highest fidelity.
        """
        if type(X) is not list:
            nlevel = 1
            X = [X]
        else:
            nlevel = len(X)

        if type(y) is not list:
            y = [y]

        if len(X) != len(y):
            raise ValueError("X and y must have the same length.")

        n_samples = np.zeros(nlevel, dtype=int)
        n_features = np.zeros(nlevel, dtype=int)
        n_samples_y = np.zeros(nlevel, dtype=int)
        for i in range(nlevel):
            n_samples[i], n_features[i] = X[i].shape
            if i > 1 and n_features[i] != n_features[i - 1]:
                raise ValueError("All X must have the same number of columns.")
            y[i] = np.asarray(y[i]).ravel()[:, np.newaxis]
            n_samples_y[i] = y[i].shape[0]
            if n_samples[i] != n_samples_y[i]:
                raise ValueError("X and y must have the same number of rows.")

        self.n_features = n_features[0]

        if type(self.theta) is not list:
            self.theta = nlevel * [self.theta]
        elif len(self.theta) != nlevel:
            raise ValueError("theta must be a list of %d element(s)." % nlevel)

        if type(self.theta0) is not list:
            self.theta0 = nlevel * [self.theta0]
        elif len(self.theta0) != nlevel:
            raise ValueError("theta0 must be a list of %d elements." % nlevel)

        if type(self.thetaL) is not list:
            self.thetaL = nlevel * [self.thetaL]
        elif len(self.thetaL) != nlevel:
            raise ValueError("thetaL must be a list of %d elements." % nlevel)

        if type(self.thetaU) is not list:
            self.thetaU = nlevel * [self.thetaU]
        elif len(self.thetaU) != nlevel:
            raise ValueError("thetaU must be a list of %d elements." % nlevel)

        self.nlevel = nlevel
        self.X = X[:]
        self.y = y[:]
        self.n_samples = n_samples

        return

    def _check_params(self):
        """
        Perform sanity checks on all parameters.
        """
        # Check regression model
        if not callable(self.regr):
            if self.regr in self._regression_types:
                self.regr = self._regression_types[self.regr]
            else:
                raise ValueError("regr should be one of %s or callable, "
                                 "%s was given."
                                 % (self._regression_types.keys(), self.regr))

        # Check rho regression model
        if not callable(self.rho_regr):
            if self.rho_regr in self._regression_types:
                self.rho_regr = self._regression_types[self.rho_regr]
            else:
                raise ValueError("rho_regr should be one of %s or callable, "
                                 "%s was given."
                                 % (self._regression_types.keys(), self.rho_regr))

        for i in range(self.nlevel):
            # Check correlation parameters
            if self.theta[i] is not None:
                self.theta[i] = array2d(self.theta[i])
                if np.any(self.theta[i] <= 0):
                    raise ValueError("theta0 must be strictly positive.")

            if self.theta0[i] is not None:
                self.theta0[i] = array2d(self.theta0[i])
                if np.any(self.theta0[i] <= 0):
                    raise ValueError("theta0 must be strictly positive.")
            else:
                self.theta0[i] = array2d(self.n_features * [THETA0_DEFAULT])

            lth = self.theta0[i].size

            if self.thetaL[i] is not None:
                self.thetaL[i] = array2d(self.thetaL[i])
                if self.thetaL[i].size != lth:
                    raise ValueError("theta0 and thetaL must have the "
                                     "same length.")
            else:
                self.thetaL[i] = array2d(self.n_features * [THETAL_DEFAULT])

            if self.thetaU[i] is not None:
                self.thetaU[i] = array2d(self.thetaU[i])
                if self.thetaU[i].size != lth:
                    raise ValueError("theta0 and thetaU must have the "
                                     "same length.")
            else:
                self.thetaU[i] = array2d(self.n_features * [THETAU_DEFAULT])

            if np.any(self.thetaL[i] <= 0) or np.any(self.thetaU[i] < self.thetaL[i]):
                raise ValueError("The bounds must satisfy O < thetaL <= "
                                 "thetaU.")

        return


class MultiFiCoKrigingSurrogate(MultiFiSurrogateModel):
    """
    OpenMDAO adapter of multi-fidelity recursive cokriging method described in [LeGratiet2013].

    See MultiFiCoKriging class.

    Attributes
    ----------
    initial_range : float
        Initial range for the optimizer.
    model : MultiFiCoKriging
        Contains MultiFiCoKriging surrogate.
    tolerance : float
        Optimizer terminates when the tolerance tol is reached.
    """

    def __init__(self, regr='constant', rho_regr='constant',
                 theta=None, theta0=None, thetaL=None, thetaU=None,
                 tolerance=TOLERANCE_DEFAULT, initial_range=INITIAL_RANGE_DEFAULT):
        """
        Initialize all attributes.

        Parameters
        ----------
        regr : string or callable, optional
            A regression function returning an array of outputs of the linear
            regression functional basis for Universal Kriging purpose.
            regr is assumed to be the same for all levels of code.
            Default assumes a simple constant regression trend.
            Available built-in regression models are:
            'constant', 'linear'
        rho_regr : string or callable, optional
            A regression function returning an array of outputs of the linear
            regression functional basis. Defines the regression function for the
            autoregressive parameter rho.
            rho_regr is assumed to be the same for all levels of code.
            Default assumes a simple constant regression trend.
            Available built-in regression models are:
            'constant', 'linear'
        theta : double, array_like or list, optional
            Value of correlation parameters if they are known; no optimization is run.
            Default is None, so that optimization is run.
            if double: value is replicated for all features and all levels.
            if array_like: an array with shape (n_features, ) for
            isotropic calculation. It is replicated for all levels.
            if list: a list of nlevel arrays specifying value for each level
        theta0 : double, array_like or list, optional
            Starting point for the maximum likelihood estimation of the
            best set of parameters.
            Default is None and meaning use of the default 0.5*np.ones(n_features)
            if double: value is replicated for all features and all levels.
            if array_like: an array with shape (n_features, ) for
            isotropic calculation. It is replicated for all levels.
            if list: a list of nlevel arrays specifying value for each level
        thetaL : double, array_like or list, optional
            Lower bound on the autocorrelation parameters for maximum
            likelihood estimation.
            Default is None meaning use of the default 1e-5*np.ones(n_features).
            if double: value is replicated for all features and all levels.
            if array_like: An array with shape matching theta0's. It is replicated
            for all levels of code.
            if list: a list of nlevel arrays specifying value for each level
        thetaU : double, array_like or list, optional
            Upper bound on the autocorrelation parameters for maximum
            likelihood estimation.
            Default is None meaning use of default value 50*np.ones(n_features).
            if double: value is replicated for all features and all levels.
            if array_like: An array with shape matching theta0's. It is replicated
            for all levels of code.
            if list: a list of nlevel arrays specifying value for each level
        tolerance : float
            Optimizer terminates when the tolerance tol is reached.
        initial_range : float
            Initial range for the optimizer.
        """
        super(MultiFiCoKrigingSurrogate, self).__init__()

        self.tolerance = tolerance
        self.initial_range = initial_range
        self.model = MultiFiCoKriging(regr=regr, rho_regr=rho_regr, theta=theta,
                                      theta0=theta0, thetaL=thetaL, thetaU=thetaU)

    def predict(self, new_x):
        """
        Calculate a predicted value of the response based on the current trained model.

        Parameters
        ----------
        new_x : array_like
            An array with shape (n_eval, n_features) giving the point(s) at
            which the prediction(s) should be made.

        Returns
        -------
        array_like
            An array with shape (n_eval, ) with the Best Linear Unbiased
            Prediction at X. If all_levels is set to True, an array
            with shape (n_eval, nlevel) giving the BLUP for all levels.

        array_like
            An array with shape (n_eval, ) with the square root of the Mean Squared Error at X.
        """
        Y_pred, MSE = self.model.predict([new_x])
        return Y_pred, np.sqrt(np.abs(MSE))

    def train_multifi(self, X, Y):
        """
        Train the surrogate model with the given set of inputs and outputs.

        Parameters
        ----------
        X : array_like
            An array with shape (n_samples_X, n_features) with the input at which observations
            were made.
        Y : array_like
            An array with shape (n_samples_X, n_features) with the observations of the scalar
            output to be predicted.
        """
        X, Y = self._fit_adapter(X, Y)
        self.model.fit(X, Y, tol=self.tolerance,
                       initial_range=self.initial_range)

    def _fit_adapter(self, X, Y):
        """
        Manage special case with one fidelity.

        where can be called as [[xval1],[xval2]] instead of [[[xval1],[xval2]]]
        we detect if shape(X[0]) is like (m,) instead of (m, n)

        Parameters
        ----------
        X : array_like
            An array with shape (n_samples_X, n_features)
        Y : array_like
            An array with shape (n_samples_Y, n_features)

        Returns
        -------
        list of double array_like elements
            A list of arrays with the input at which observations were made, from lowest
            fidelity to highest fidelity. Designs must be nested
            with X[i] = np.vstack([..., X[i+1])
        list of double array_like elements
            A list of arrays with the observations of the scalar output to be predicted,
            from lowest fidelity to highest fidelity.
        """
        if len(np.shape(np.array(X[0]))) == 1:
            X = [X]
            Y = [Y]

        X = [np.array(x) for x in reversed(X)]
        Y = [np.array(y) for y in reversed(Y)]
        return (X, Y)


class FloatMultiFiCoKrigingSurrogate(MultiFiCoKrigingSurrogate):
    """
    Predictions are returned as the mean of the NormalDistribution predicted by base class model.
    """

    def predict(self, new_x):
        """
        Calculate a predicted value of the response based on the current trained model.

        Parameters
        ----------
        new_x : array_like
            An array with shape (n_eval, n_features) giving the point(s) at
            which the prediction(s) should be made.

        Returns
        -------
        float
            New predicted value.
        """
        dist = super(FloatMultiFiCoKrigingSurrogate, self).predict(new_x)
        return dist.mu


if __name__ == "__main__":
    import doctest
    doctest.testmod()
